#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from tempest.lib import exceptions

from cinderclient.tests.functional import base


@ddt.ddt
class CinderVolumeNegativeTests(base.ClientTestBase):
    """Check of cinder volume create commands."""

    @ddt.data(
        ('', (r'Size is a required parameter')),
        ('-1', (r'Invalid input for field/attribute size')),
        ('0', (r"Invalid input for field/attribute size")),
        ('size', (r'invalid int value')),
        ('0.2', (r'invalid int value')),
        ('2 GB', (r'unrecognized arguments')),
        ('999999999', (r'VolumeSizeExceedsAvailableQuota')),
    )
    @ddt.unpack
    def test_volume_create_with_incorrect_size(self, value, ex_text):
        self.assertRaisesRegex(exceptions.CommandFailed,
                               ex_text, self.object_create,
                               'volume', params=value)


class CinderVolumeTests(base.ClientTestBase):
    """Check of cinder volume create commands."""
    def setUp(self):
        super(CinderVolumeTests, self).setUp()
        self.volume = self.object_create('volume', params='1')

    def test_volume_create_from_snapshot(self):
        """Test steps:

        1) create volume in Setup()
        2) create snapshot
        3) create volume from snapshot
        4) check that volume from snapshot has been successfully created
        """
        snapshot = self.object_create('snapshot', params=self.volume['id'])
        volume_from_snapshot = self.object_create('volume',
                                           params='--snapshot-id {0} 1'.
                                           format(snapshot['id']))
        self.object_delete('snapshot', snapshot['id'])
        self.check_object_deleted('snapshot', snapshot['id'])
        cinder_list = self.cinder('list')
        self.assertIn(volume_from_snapshot['id'], cinder_list)

    def test_volume_create_from_volume(self):
        """Test steps:

        1) create volume in Setup()
        2) create volume from volume
        3) check that volume from volume has been successfully created
        """
        volume_from_volume = self.object_create('volume',
                                         params='--source-volid {0} 1'.
                                         format(self.volume['id']))
        cinder_list = self.cinder('list')
        self.assertIn(volume_from_volume['id'], cinder_list)


class CinderVolumeTestsWithParameters(base.ClientTestBase):
    """Check of cinder volume create commands with parameters."""
    def test_volume_create_description(self):
        """Test steps:

        1) create volume with description
        2) check that volume has right description
        """
        volume_description = 'test_description'
        volume = self.object_create('volume',
                                    params='--description {0} 1'.
                                    format(volume_description))
        self.assertEqual(volume_description, volume['description'])

    def test_volume_create_metadata(self):
        """Test steps:

        1) create volume with metadata
        2) check that metadata complies entered
        """
        volume = self.object_create(
            'volume', params='--metadata test_metadata=test_date 1')
        self.assertEqual(str({'test_metadata': 'test_date'}),
                         volume['metadata'])



# Copyright (c) 2016 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import api_versions
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes
from cinderclient.v3 import services


class ServicesTest(utils.TestCase):

    def test_list_services_with_cluster_info(self):
        cs = fakes.FakeClient(api_version=api_versions.APIVersion('3.7'))
        services_list = cs.services.list()
        cs.assert_called('GET', '/os-services')
        self.assertEqual(3, len(services_list))
        for service in services_list:
            self.assertIsInstance(service, services.Service)
            # Make sure cluster fields from v3.7 is present and not None
            self.assertIsNotNone(getattr(service, 'cluster'))
        self._assert_request_id(services_list)

    def test_api_version(self):
        client = fakes.FakeClient(version_header='3.0')
        svs = client.services.server_api_version()
        [self.assertIsInstance(s, services.Service) for s in svs]

    def test_set_log_levels(self):
        expected = {'level': 'debug', 'binary': 'cinder-api',
                    'server': 'host1', 'prefix': 'sqlalchemy.'}

        cs = fakes.FakeClient(version_header='3.32')
        cs.services.set_log_levels(expected['level'], expected['binary'],
                                   expected['server'], expected['prefix'])

        cs.assert_called('PUT', '/os-services/set-log', body=expected)

    def test_get_log_levels(self):
        expected = {'binary': 'cinder-api', 'server': 'host1',
                    'prefix': 'sqlalchemy.'}

        cs = fakes.FakeClient(version_header='3.32')
        result = cs.services.get_log_levels(expected['binary'],
                                            expected['server'],
                                            expected['prefix'])

        cs.assert_called('PUT', '/os-services/get-log', body=expected)
        expected = [services.LogLevel(cs.services,
                                      {'binary': 'cinder-api', 'host': 'host1',
                                       'prefix': 'prefix1', 'level': 'DEBUG'},
                                      loaded=True),
                    services.LogLevel(cs.services,
                                      {'binary': 'cinder-api', 'host': 'host1',
                                       'prefix': 'prefix2', 'level': 'INFO'},
                                      loaded=True),
                    services.LogLevel(cs.services,
                                      {'binary': 'cinder-volume',
                                       'host': 'host@backend#pool',
                                       'prefix': 'prefix3',
                                       'level': 'WARNING'},
                                      loaded=True),
                    services.LogLevel(cs.services,
                                      {'binary': 'cinder-volume',
                                       'host': 'host@backend#pool',
                                       'prefix': 'prefix4', 'level': 'ERROR'},
                                      loaded=True)]
        # Since it will be sorted by the prefix we can compare them directly
        self.assertListEqual(expected, result)

    def test_list_services_with_backend_state(self):
        cs = fakes.FakeClient(api_version=api_versions.APIVersion('3.49'))
        services_list = cs.services.list()
        cs.assert_called('GET', '/os-services')
        self.assertEqual(3, len(services_list))
        for service in services_list:
            self.assertIsInstance(service, services.Service)
            # Make sure backend_state fields from v3.49 is present and not
            # None
            if service.binary == 'cinder-volume':
                self.assertIsNotNone(getattr(service, 'backend_state',
                                             None))
        self._assert_request_id(services_list)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
import time

import prettytable

from cinderclient import exceptions
from cinderclient import utils

_quota_resources = ['volumes', 'snapshots', 'gigabytes',
                    'backups', 'backup_gigabytes',
                    'per_volume_gigabytes', 'groups', ]
_quota_infos = ['Type', 'In_use', 'Reserved', 'Limit', 'Allocated']


def _print(pt, order):
    print(pt.get_string(sortby=order))


def _pretty_format_dict(data_dict):
    formatted_data = []

    for k in sorted(data_dict):
        formatted_data.append("%s : %s" % (k, data_dict[k]))

    return "\n".join(formatted_data)


def print_list(objs, fields, exclude_unavailable=False, formatters=None,
               sortby_index=0):
    '''Prints a list of objects.

    @param objs: Objects to print
    @param fields: Fields on each object to be printed
    @param exclude_unavailable: Boolean to decide if unavailable fields are
                                removed
    @param formatters: Custom field formatters
    @param sortby_index: Results sorted against the key in the fields list at
                         this index; if None then the object order is not
                         altered
    '''
    formatters = formatters or {}
    mixed_case_fields = ['serverId']
    removed_fields = []
    rows = []

    for o in objs:
        row = []
        for field in fields:
            if field in removed_fields:
                continue
            if field in formatters:
                row.append(formatters[field](o))
            else:
                if field in mixed_case_fields:
                    field_name = field.replace(' ', '_')
                else:
                    field_name = field.lower().replace(' ', '_')
                if isinstance(o, dict) and field in o:
                    data = o[field]
                else:
                    if not hasattr(o, field_name) and exclude_unavailable:
                        removed_fields.append(field)
                        continue
                    else:
                        data = getattr(o, field_name, '')
                if data is None:
                    data = '-'
                if isinstance(data, str) and "\r" in data:
                    data = data.replace("\r", " ")
                row.append(data)
        rows.append(row)

    for f in removed_fields:
        fields.remove(f)

    pt = prettytable.PrettyTable((f for f in fields), caching=False)
    pt.align = 'l'
    for row in rows:
        count = 0
        # Converts unicode values in dictionary to string
        for part in row:
            count = count + 1
            if isinstance(part, dict):
                row[count - 1] = part
        pt.add_row(row)

    if sortby_index is None:
        order_by = None
    else:
        order_by = fields[sortby_index]
    _print(pt, order_by)


def print_dict(d, property="Property", formatters=None):
    pt = prettytable.PrettyTable([property, 'Value'], caching=False)
    pt.align = 'l'
    formatters = formatters or {}

    for r in d.items():
        r = list(r)

        if r[0] in formatters:
            if isinstance(r[1], dict):
                r[1] = _pretty_format_dict(r[1])
        if isinstance(r[1], str) and "\r" in r[1]:
            r[1] = r[1].replace("\r", " ")
        pt.add_row(r)
    _print(pt, property)


def print_volume_image(image_resp_tuple):
    # image_resp_tuple = tuple (response, body)
    image = image_resp_tuple[1]
    vt = image['os-volume_upload_image'].get('volume_type')
    if vt is not None:
        image['os-volume_upload_image']['volume_type'] = vt.get('name')
    print_dict(image['os-volume_upload_image'])


def poll_for_status(poll_fn, obj_id, action, final_ok_states,
                    poll_period=5, show_progress=True):
    """Blocks while an action occurs. Periodically shows progress."""
    def print_progress(progress):
        if show_progress:
            msg = ('\rInstance %(action)s... %(progress)s%% complete'
                   % dict(action=action, progress=progress))
        else:
            msg = '\rInstance %(action)s...' % dict(action=action)

        sys.stdout.write(msg)
        sys.stdout.flush()

    print()
    while True:
        obj = poll_fn(obj_id)
        status = obj.status.lower()
        progress = getattr(obj, 'progress', None) or 0
        if status in final_ok_states:
            print_progress(100)
            print("\nFinished")
            break
        elif status == "error":
            print("\nError %(action)s instance" % {'action': action})
            break
        else:
            print_progress(progress)
            time.sleep(poll_period)


def find_volume_snapshot(cs, snapshot):
    """Gets a volume snapshot by name or ID."""
    return utils.find_resource(cs.volume_snapshots, snapshot)


def find_vtype(cs, vtype):
    """Gets a volume type by name or ID."""
    return utils.find_resource(cs.volume_types, vtype)


def find_gtype(cs, gtype):
    """Gets a group type by name or ID."""
    return utils.find_resource(cs.group_types, gtype)


def find_backup(cs, backup):
    """Gets a backup by name or ID."""
    return utils.find_resource(cs.backups, backup)


def find_consistencygroup(cs, consistencygroup):
    """Gets a consistency group by name or ID."""
    return utils.find_resource(cs.consistencygroups, consistencygroup)


def find_group(cs, group, **kwargs):
    """Gets a group by name or ID."""
    kwargs['is_group'] = True
    return utils.find_resource(cs.groups, group, **kwargs)


def find_cgsnapshot(cs, cgsnapshot):
    """Gets a cgsnapshot by name or ID."""
    return utils.find_resource(cs.cgsnapshots, cgsnapshot)


def find_group_snapshot(cs, group_snapshot):
    """Gets a group_snapshot by name or ID."""
    return utils.find_resource(cs.group_snapshots, group_snapshot)


def find_transfer(cs, transfer):
    """Gets a transfer by name or ID."""
    return utils.find_resource(cs.transfers, transfer)


def find_qos_specs(cs, qos_specs):
    """Gets a qos specs by ID."""
    return utils.find_resource(cs.qos_specs, qos_specs)


def find_message(cs, message):
    """Gets a message by ID."""
    return utils.find_resource(cs.messages, message)


def print_volume_snapshot(snapshot):
    print_dict(snapshot._info)


def translate_keys(collection, convert):
    for item in collection:
        keys = item.__dict__
        for from_key, to_key in convert:
            if from_key in keys and to_key not in keys:
                setattr(item, to_key, item._info[from_key])


def translate_volume_keys(collection):
    convert = [('volumeType', 'volume_type'),
               ('os-vol-tenant-attr:tenant_id', 'tenant_id')]
    translate_keys(collection, convert)


def translate_volume_snapshot_keys(collection):
    convert = [('volumeId', 'volume_id')]
    translate_keys(collection, convert)


def translate_availability_zone_keys(collection):
    convert = [('zoneName', 'name'), ('zoneState', 'status')]
    translate_keys(collection, convert)


def extract_filters(args):
    filters = {}
    for f in args:
        if '=' in f:
            (key, value) = f.split('=', 1)
            if value.startswith('{') and value.endswith('}'):
                value = _build_internal_dict(value[1:-1])
            filters[key] = value
        else:
            print("WARNING: Ignoring the filter %s while showing result." % f)

    return filters


def _build_internal_dict(content):
    result = {}
    for pair in content.split(','):
        k, v = pair.split(':', 1)
        result.update({k.strip(): v.strip()})
    return result


def extract_metadata(args, type='user_metadata'):
    metadata = {}
    if type == 'image_metadata':
        args_metadata = args.image_metadata
    else:
        args_metadata = args.metadata
    for metadatum in args_metadata:
        # unset doesn't require a val, so we have the if/else
        if '=' in metadatum:
            (key, value) = metadatum.split('=', 1)
        else:
            key = metadatum
            value = None

        metadata[key] = value
    return metadata


def print_volume_type_list(vtypes):
    print_list(vtypes, ['ID', 'Name', 'Description', 'Is_Public'])


def print_group_type_list(gtypes):
    print_list(gtypes, ['ID', 'Name', 'Description'])


def print_resource_filter_list(filters):
    formatter = {'Filters': lambda resource: ', '.join(resource.filters)}
    print_list(filters, ['Resource', 'Filters'], formatters=formatter)


def quota_show(quotas):
    quotas_info_dict = quotas._info
    quota_dict = {}
    for resource in quotas_info_dict.keys():
        good_name = False
        for name in _quota_resources:
            if resource.startswith(name):
                good_name = True
        if not good_name:
            continue
        quota_dict[resource] = getattr(quotas, resource, None)
    print_dict(quota_dict)


def quota_usage_show(quotas):
    quota_list = []
    quotas_info_dict = quotas._info
    for resource in quotas_info_dict.keys():
        good_name = False
        for name in _quota_resources:
            if resource.startswith(name):
                good_name = True
        if not good_name:
            continue
        quota_info = getattr(quotas, resource, None)
        quota_info['Type'] = resource
        quota_info = dict((k.capitalize(), v) for k, v in quota_info.items())
        quota_list.append(quota_info)
    print_list(quota_list, _quota_infos)


def quota_update(manager, identifier, args):
    updates = {}
    for resource in _quota_resources:
        val = getattr(args, resource, None)
        if val is not None:
            if args.volume_type:
                resource = resource + '_%s' % args.volume_type
            updates[resource] = val

    if updates:
        skip_validation = getattr(args, 'skip_validation', True)
        if not skip_validation:
            updates['skip_validation'] = skip_validation
        quota_show(manager.update(identifier, **updates))
    else:
        msg = 'Must supply at least one quota field to update.'
        raise exceptions.ClientException(code=1, message=msg)


def find_volume_type(cs, vtype):
    """Gets a volume type by name or ID."""
    return utils.find_resource(cs.volume_types, vtype)


def find_group_type(cs, gtype):
    """Gets a group type by name or ID."""
    return utils.find_resource(cs.group_types, gtype)


def print_volume_encryption_type_list(encryption_types):
    """
    Lists volume encryption types.

    :param encryption_types: a list of :class: VolumeEncryptionType instances
    """
    print_list(encryption_types, ['Volume Type ID', 'Provider',
                                  'Cipher', 'Key Size',
                                  'Control Location'])


def print_qos_specs(qos_specs):
    # formatters defines field to be converted from unicode to string
    print_dict(qos_specs._info, formatters=['specs'])


def print_qos_specs_list(q_specs):
    print_list(q_specs, ['ID', 'Name', 'Consumer', 'specs'])


def print_qos_specs_and_associations_list(q_specs):
    print_list(q_specs, ['ID', 'Name', 'Consumer', 'specs'])


def print_associations_list(associations):
    print_list(associations, ['Association_Type', 'Name', 'ID'])


def _poll_for_status(poll_fn, obj_id, info, action, final_ok_states,
                     timeout_period, global_request_id=None, messages=None,
                     poll_period=2, status_field="status"):
    """Block while an action is being performed."""
    time_elapsed = 0
    while True:
        time.sleep(poll_period)
        time_elapsed += poll_period
        obj = poll_fn(obj_id)
        status = getattr(obj, status_field)
        info[status_field] = status
        if status:
            status = status.lower()

        if status in final_ok_states:
            break
        elif status == "error":
            print_dict(info)
            if global_request_id:
                search_opts = {
                    'request_id': global_request_id
                    }
                message_list = messages.list(search_opts=search_opts)
                try:
                    fault_msg = message_list[0].user_message
                except IndexError:
                    fault_msg = "Unknown error. Operation failed."
                raise exceptions.ResourceInErrorState(obj, fault_msg)
        elif time_elapsed == timeout_period:
            print_dict(info)
            raise exceptions.TimeoutException(obj, action)



# Copyright (c) 2016 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import api_versions
from cinderclient import exceptions as exc
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes
from cinderclient.v3 import volume_backups_restore


class VolumesTest(utils.TestCase):

    def test_update(self):
        cs = fakes.FakeClient(api_version=api_versions.APIVersion('3.9'))
        b = cs.backups.get('1234')
        backup = b.update(name='new-name')
        cs.assert_called(
            'PUT', '/backups/1234',
            {'backup': {'name': 'new-name'}})
        self._assert_request_id(backup)

    def test_pre_version(self):
        cs = fakes.FakeClient(api_version=api_versions.APIVersion('3.8'))
        b = cs.backups.get('1234')
        self.assertRaises(exc.VersionNotFoundForAPIMethod,
                          b.update, name='new-name')

    def test_restore(self):
        cs = fakes.FakeClient(api_version=api_versions.APIVersion('3.0'))
        backup_id = '76a17945-3c6f-435c-975b-b5685db10b62'
        info = cs.restores.restore(backup_id)
        cs.assert_called('POST', '/backups/%s/restore' % backup_id)
        self.assertIsInstance(info,
                              volume_backups_restore.VolumeBackupsRestore)
        self._assert_request_id(info)

    def test_restore_with_name(self):
        cs = fakes.FakeClient(api_version=api_versions.APIVersion('3.0'))
        backup_id = '76a17945-3c6f-435c-975b-b5685db10b62'
        name = 'restore_vol'
        info = cs.restores.restore(backup_id, name=name)
        expected_body = {'restore': {'volume_id': None, 'name': name}}
        cs.assert_called('POST', '/backups/%s/restore' % backup_id,
                         body=expected_body)
        self.assertIsInstance(info,
                              volume_backups_restore.VolumeBackupsRestore)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import logging
import re

from oslo_utils import strutils

from cinderclient._i18n import _
from cinderclient import exceptions
from cinderclient import utils

LOG = logging.getLogger(__name__)


# key is unsupported version, value is appropriate supported alternative
REPLACEMENT_VERSIONS = {"1": "3", "2": "3"}
MAX_VERSION = "3.71"
MIN_VERSION = "3.0"

_SUBSTITUTIONS = {}

_type_error_msg = "'%(other)s' should be an instance of '%(cls)s'"


class APIVersion(object):
    """This class represents an API Version with convenience
    methods for manipulation and comparison of version
    numbers that we need to do to implement microversions.
    """

    def __init__(self, version_str=None):
        """Create an API version object."""
        self.ver_major = 0
        self.ver_minor = 0

        if version_str is not None:
            match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0|latest)$", version_str)
            if match:
                self.ver_major = int(match.group(1))
                if match.group(2) == "latest":
                    # NOTE(andreykurilin): Infinity allows to easily determine
                    # latest version and doesn't require any additional checks
                    # in comparison methods.
                    self.ver_minor = float("inf")
                else:
                    self.ver_minor = int(match.group(2))
            else:
                msg = (_("Invalid format of client version '%s'. "
                       "Expected format 'X.Y', where X is a major part and Y "
                       "is a minor part of version.") % version_str)
                raise exceptions.UnsupportedVersion(msg)

    def __str__(self):
        """Debug/Logging representation of object."""
        if self.is_latest():
            return "Latest API Version Major: %s" % self.ver_major
        return ("API Version Major: %s, Minor: %s"
                % (self.ver_major, self.ver_minor))

    def __repr__(self):
        if self:
            return "<APIVersion: %s>" % self.get_string()
        return "<APIVersion: null>"

    def __bool__(self):
        return self.ver_major != 0 or self.ver_minor != 0

    __nonzero__ = __bool__

    def is_latest(self):
        return self.ver_minor == float("inf")

    def __lt__(self, other):
        if not isinstance(other, APIVersion):
            raise TypeError(_type_error_msg % {"other": other,
                                               "cls": self.__class__})

        return ((self.ver_major, self.ver_minor) <
                (other.ver_major, other.ver_minor))

    def __eq__(self, other):
        if not isinstance(other, APIVersion):
            raise TypeError(_type_error_msg % {"other": other,
                                               "cls": self.__class__})

        return ((self.ver_major, self.ver_minor) ==
                (other.ver_major, other.ver_minor))

    def __gt__(self, other):
        if not isinstance(other, APIVersion):
            raise TypeError(_type_error_msg % {"other": other,
                                               "cls": self.__class__})

        return ((self.ver_major, self.ver_minor) >
                (other.ver_major, other.ver_minor))

    def __le__(self, other):
        return self < other or self == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return self > other or self == other

    def matches(self, min_version, max_version=None):
        """Returns whether the version object represents a version
        greater than or equal to the minimum version and less than
        or equal to the maximum version.

        :param min_version: Minimum acceptable version.
        :param max_version: Maximum acceptable version.
        :returns: boolean

        If min_version is null then there is no minimum limit.
        If max_version is null then there is no maximum limit.
        If self is null then raise ValueError
        """

        if not self:
            raise ValueError("Null APIVersion doesn't support 'matches'.")

        if isinstance(min_version, str):
            min_version = APIVersion(version_str=min_version)
        if isinstance(max_version, str):
            max_version = APIVersion(version_str=max_version)

        # This will work when they are None and when they are version 0.0
        if not min_version and not max_version:
            return True

        if not max_version:
            return min_version <= self
        if not min_version:
            return self <= max_version
        return min_version <= self <= max_version

    def get_string(self):
        """Converts object to string representation which if used to create
        an APIVersion object results in the same version.
        """
        if not self:
            raise ValueError("Null APIVersion cannot be converted to string.")
        elif self.is_latest():
            return "%s.%s" % (self.ver_major, "latest")
        return "%s.%s" % (self.ver_major, self.ver_minor)

    def get_major_version(self):
        return "%s" % self.ver_major


class VersionedMethod(object):

    def __init__(self, name, start_version, end_version, func):
        """Versioning information for a single method

        :param name: Name of the method
        :param start_version: Minimum acceptable version
        :param end_version: Maximum acceptable_version
        :param func: Method to call

        Minimum and maximums are inclusive
        """
        self.name = name
        self.start_version = start_version
        self.end_version = end_version
        self.func = func

    def __str__(self):
        return ("Version Method %s: min: %s, max: %s"
                % (self.name, self.start_version, self.end_version))

    def __repr__(self):
        return "<VersionedMethod %s>" % self.name


def get_available_major_versions():
    # NOTE: the discovery code previously here assumed that if a v2
    # module exists, it must contain a client.  This will be False
    # during the transition period when the v2 client is removed but
    # we are still using other classes in that module.  Right now there's
    # only one client version available, so we simply hard-code it.
    return ['3']


def check_major_version(api_version):
    """Checks major part of ``APIVersion`` obj is supported.

    :raises cinderclient.exceptions.UnsupportedVersion: if major part is not
                                                      supported
    """
    available_versions = get_available_major_versions()
    if (api_version and str(api_version.ver_major) not in available_versions):
        if len(available_versions) == 1:
            msg = ("Invalid client version '%(version)s'. "
                   "Major part should be '%(major)s'") % {
                "version": api_version.get_string(),
                "major": available_versions[0]}
        else:
            msg = ("Invalid client version '%(version)s'. "
                   "Major part must be one of: '%(major)s'") % {
                "version": api_version.get_string(),
                "major": ", ".join(available_versions)}
        raise exceptions.UnsupportedVersion(msg)


def get_api_version(version_string):
    """Returns checked APIVersion object"""
    version_string = str(version_string)
    if version_string in REPLACEMENT_VERSIONS:
        LOG.warning("Version %(old)s is not supported, use "
                    "supported version %(now)s instead.",
                   {"old": version_string,
                    "now": REPLACEMENT_VERSIONS[version_string]})
    if strutils.is_int_like(version_string):
        version_string = "%s.0" % version_string

    api_version = APIVersion(version_string)
    check_major_version(api_version)
    return api_version


def _get_server_version_range(client):
    try:
        versions = client.services.server_api_version()
    except AttributeError:
        # Wrong client was used, translate to something helpful.
        raise exceptions.UnsupportedVersion(
            _('Invalid client version %s to get server version range. Only '
              'the v3 client is supported for this operation.') %
            client.version)

    if not versions:
        msg = _("Server does not support microversions. You cannot use this "
                "version of the cinderclient with the requested server. "
                "Try using a cinderclient version less than 8.0.0.")
        raise exceptions.UnsupportedVersion(msg)

    for version in versions:
        if '3.' in version.version:
            return APIVersion(version.min_version), APIVersion(version.version)

    # if we're still here, there's nothing we understand in the versions
    msg = _("You cannot use this version of the cinderclient with the "
            "requested server.")
    raise exceptions.UnsupportedVersion(msg)


def get_highest_version(client):
    """Queries the server version info and returns highest supported
    microversion

    :param client: client object
    :returns: APIVersion
    """
    server_start_version, server_end_version = _get_server_version_range(
        client)
    return server_end_version


def discover_version(client, requested_version):
    """Checks ``requested_version`` and returns the most recent version
    supported by both the API and the client.

    :param client: client object
    :param requested_version: requested version represented by APIVersion obj
    :returns: APIVersion
    """

    server_start_version, server_end_version = _get_server_version_range(
        client)

    _validate_server_version(server_start_version, server_end_version)

    # get the highest version the server can handle relative to the
    # requested version
    valid_version = _validate_requested_version(
        requested_version,
        server_start_version,
        server_end_version)

    # see if we need to downgrade for the client
    client_max = APIVersion(MAX_VERSION)
    if client_max < valid_version:
        msg = _("Requested version %(requested_version)s is "
                "not supported. Downgrading requested version "
                "to %(actual_version)s.")
        LOG.debug(msg, {
            "requested_version": requested_version,
            "actual_version": client_max})
        valid_version = client_max

    return valid_version


def _validate_requested_version(requested_version,
                                server_start_version,
                                server_end_version):
    """Validates the requested version.

    Checks 'requested_version' is within the min/max range supported by the
    server. If 'requested_version' is not within range then attempts to
    downgrade to 'server_end_version'. Otherwise an UnsupportedVersion
    exception is thrown.

    :param requested_version: requestedversion represented by APIVersion obj
    :param server_start_version: APIVersion object representing server min
    :param server_end_version: APIVersion object representing server max
    """
    valid_version = requested_version
    if not requested_version.matches(server_start_version, server_end_version):
        if server_end_version <= requested_version:
            if (APIVersion(MIN_VERSION) <= server_end_version and
                    server_end_version <= APIVersion(MAX_VERSION)):
                msg = _("Requested version %(requested_version)s is "
                        "not supported. Downgrading requested version "
                        "to %(server_end_version)s.")
                LOG.debug(msg, {
                    "requested_version": requested_version,
                    "server_end_version": server_end_version})
            valid_version = server_end_version
        else:
            raise exceptions.UnsupportedVersion(
                _("The specified version isn't supported by server. The valid "
                  "version range is '%(min)s' to '%(max)s'") % {
                    "min": server_start_version.get_string(),
                    "max": server_end_version.get_string()})

    return valid_version


def _validate_server_version(server_start_version, server_end_version):
    """Validates the server version.

    Checks that the 'server_end_version' is greater than the minimum version
    supported by the client. Then checks that the 'server_start_version' is
    less than the maximum version supported by the client.

    :param server_start_version:
    :param server_end_version:
    :return:
    """
    if APIVersion(MIN_VERSION) > server_end_version:
        raise exceptions.UnsupportedVersion(
            _("Server's version is too old. The client's valid version range "
              "is '%(client_min)s' to '%(client_max)s'. The server valid "
              "version range is '%(server_min)s' to '%(server_max)s'.") % {
                  'client_min': MIN_VERSION,
                  'client_max': MAX_VERSION,
                  'server_min': server_start_version.get_string(),
                  'server_max': server_end_version.get_string()})
    elif APIVersion(MAX_VERSION) < server_start_version:
        raise exceptions.UnsupportedVersion(
            _("Server's version is too new. The client's valid version range "
              "is '%(client_min)s' to '%(client_max)s'. The server valid "
              "version range is '%(server_min)s' to '%(server_max)s'.") % {
                  'client_min': MIN_VERSION,
                  'client_max': MAX_VERSION,
                  'server_min': server_start_version.get_string(),
                  'server_max': server_end_version.get_string()})


def update_headers(headers, api_version):
    """Set 'OpenStack-API-Version' header if api_version is not
       null
    """

    if api_version and api_version.ver_minor != 0:
        headers["OpenStack-API-Version"] = "volume " + api_version.get_string()


def add_substitution(versioned_method):
    _SUBSTITUTIONS.setdefault(versioned_method.name, [])
    _SUBSTITUTIONS[versioned_method.name].append(versioned_method)


def get_substitutions(func_name, api_version=None):
    substitutions = _SUBSTITUTIONS.get(func_name, [])
    if api_version:
        return [m for m in substitutions
                if api_version.matches(m.start_version, m.end_version)]
    return substitutions


def wraps(start_version, end_version=None):
    start_version = APIVersion(start_version)
    if end_version:
        end_version = APIVersion(end_version)
    else:
        end_version = APIVersion("%s.latest" % start_version.ver_major)

    def decor(func):
        func.versioned = True
        name = utils.get_function_name(func)
        versioned_method = VersionedMethod(name, start_version,
                                           end_version, func)
        add_substitution(versioned_method)

        @functools.wraps(func)
        def substitution(obj, *args, **kwargs):
            methods = get_substitutions(name, obj.api_version)

            if not methods:
                raise exceptions.VersionNotFoundForAPIMethod(
                    obj.api_version.get_string(), name)

            method = max(methods, key=lambda f: f.start_version)

            return method.func(obj, *args, **kwargs)

        if hasattr(func, 'arguments'):
            for cli_args, cli_kwargs in func.arguments:
                utils.add_arg(substitution, *cli_args, **cli_kwargs)
        return substitution

    return decor



# Copyright (C) 2016 EMC Corporation.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import api_versions
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes


class AttachmentsTest(utils.TestCase):

    def test_create_attachment(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.27'))
        att = cs.attachments.create(
            'e84fda45-4de4-4ce4-8f39-fc9d3b0aa05e',
            {},
            '557ad76c-ce54-40a3-9e91-c40d21665cc3',
            'null')
        cs.assert_called('POST', '/attachments')
        self.assertEqual(fakes.fake_attachment['attachment'], att)

    def test_create_attachment_without_instance_uuid(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.27'))
        att = cs.attachments.create(
            'e84fda45-4de4-4ce4-8f39-fc9d3b0aa05e',
            {},
            None,
            'null')
        cs.assert_called('POST', '/attachments')
        self.assertEqual(
            fakes.fake_attachment_without_instance_id['attachment'], att)

    def test_complete_attachment(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.44'))
        att = cs.attachments.complete('a232e9ae')
        self.assertTrue(att.ok)



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cinderclient import base
from cinderclient import utils


class Limits(base.Resource):
    """A collection of RateLimit and AbsoluteLimit objects."""

    def __repr__(self):
        return "<Limits>"

    @property
    def absolute(self):
        for (name, value) in list(self._info['absolute'].items()):
            yield AbsoluteLimit(name, value)

    @property
    def rate(self):
        for group in self._info['rate']:
            uri = group['uri']
            regex = group['regex']
            for rate in group['limit']:
                yield RateLimit(rate['verb'], uri, regex, rate['value'],
                                rate['remaining'], rate['unit'],
                                rate['next-available'])


class RateLimit(object):
    """Data model that represents a flattened view of a single rate limit."""

    def __init__(self, verb, uri, regex, value, remain,
                 unit, next_available):
        self.verb = verb
        self.uri = uri
        self.regex = regex
        self.value = value
        self.remain = remain
        self.unit = unit
        self.next_available = next_available

    def __eq__(self, other):
        return self.uri == other.uri \
            and self.regex == other.regex \
            and self.value == other.value \
            and self.verb == other.verb \
            and self.remain == other.remain \
            and self.unit == other.unit \
            and self.next_available == other.next_available

    def __repr__(self):
        return "<RateLimit: method=%s uri=%s>" % (self.verb, self.uri)


class AbsoluteLimit(object):
    """Data model that represents a single absolute limit."""

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __eq__(self, other):
        return self.value == other.value and self.name == other.name

    def __repr__(self):
        return "<AbsoluteLimit: name=%s>" % (self.name)


class LimitsManager(base.Manager):
    """Manager object used to interact with limits resource."""

    resource_class = Limits

    def get(self, tenant_id=None):
        """Get a specific extension.

        :rtype: :class:`Limits`
        """
        opts = {}
        if tenant_id:
            opts['tenant_id'] = tenant_id

        query_string = utils.build_query_param(opts)

        return self._get("/limits%s" % query_string, "limits")



# Copyright 2016 Mirantis
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from cinderclient import api_versions
from cinderclient import exceptions
from cinderclient.tests.unit import test_utils
from cinderclient.tests.unit import utils
from cinderclient.v3 import client


@ddt.ddt
class APIVersionTestCase(utils.TestCase):
    def test_valid_version_strings(self):
        def _test_string(version, exp_major, exp_minor):
            v = api_versions.APIVersion(version)
            self.assertEqual(v.ver_major, exp_major)
            self.assertEqual(v.ver_minor, exp_minor)

        _test_string("1.1", 1, 1)
        _test_string("2.10", 2, 10)
        _test_string("5.234", 5, 234)
        _test_string("12.5", 12, 5)
        _test_string("2.0", 2, 0)
        _test_string("2.200", 2, 200)

    def test_null_version(self):
        v = api_versions.APIVersion()
        self.assertFalse(v)

    def test_not_null_version(self):
        v = api_versions.APIVersion('1.1')
        self.assertTrue(v)

    @ddt.data("2", "200", "2.1.4", "200.23.66.3", "5 .3", "5. 3", "5.03",
              "02.1", "2.001", "", " 2.1", "2.1 ")
    def test_invalid_version_strings(self, version_string):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.APIVersion, version_string)

    def test_version_comparisons(self):
        v1 = api_versions.APIVersion("2.0")
        v2 = api_versions.APIVersion("2.5")
        v3 = api_versions.APIVersion("5.23")
        v4 = api_versions.APIVersion("2.0")
        v_null = api_versions.APIVersion()

        self.assertLess(v1, v2)
        self.assertGreater(v3, v2)
        self.assertNotEqual(v1, v2)
        self.assertEqual(v1, v4)
        self.assertNotEqual(v1, v_null)
        self.assertEqual(v_null, v_null)
        self.assertRaises(TypeError, v1.__le__, "2.1")

    def test_version_matches(self):
        v1 = api_versions.APIVersion("2.0")
        v2 = api_versions.APIVersion("2.5")
        v3 = api_versions.APIVersion("2.45")
        v4 = api_versions.APIVersion("3.3")
        v5 = api_versions.APIVersion("3.23")
        v6 = api_versions.APIVersion("2.0")
        v7 = api_versions.APIVersion("3.3")
        v8 = api_versions.APIVersion("4.0")
        v_null = api_versions.APIVersion()

        self.assertTrue(v2.matches(v1, v3))
        self.assertTrue(v2.matches(v1, v_null))
        self.assertTrue(v1.matches(v6, v2))
        self.assertTrue(v4.matches(v2, v7))
        self.assertTrue(v4.matches(v_null, v7))
        self.assertTrue(v4.matches(v_null, v8))
        self.assertFalse(v1.matches(v2, v3))
        self.assertFalse(v5.matches(v2, v4))
        self.assertFalse(v2.matches(v3, v1))

        self.assertRaises(ValueError, v_null.matches, v1, v3)

    def test_get_string(self):
        v1_string = "3.23"
        v1 = api_versions.APIVersion(v1_string)
        self.assertEqual(v1_string, v1.get_string())

        self.assertRaises(ValueError,
                          api_versions.APIVersion().get_string)


class ManagerTest(utils.TestCase):
    def test_api_version(self):
        # The function manager.return_api_version has two versions,
        # when called with api version 3.1 it should return the
        # string '3.1' and when called with api version 3.2 or higher
        # it should return the string '3.2'.
        version = api_versions.APIVersion('3.1')
        api = client.Client(api_version=version)
        manager = test_utils.FakeManagerWithApi(api)
        self.assertEqual('3.1', manager.return_api_version())

        version = api_versions.APIVersion('3.2')
        api = client.Client(api_version=version)
        manager = test_utils.FakeManagerWithApi(api)
        self.assertEqual('3.2', manager.return_api_version())

        # pick up the highest version
        version = api_versions.APIVersion('3.3')
        api = client.Client(api_version=version)
        manager = test_utils.FakeManagerWithApi(api)
        self.assertEqual('3.2', manager.return_api_version())

        version = api_versions.APIVersion('3.0')
        api = client.Client(api_version=version)
        manager = test_utils.FakeManagerWithApi(api)
        # An exception will be returned here because the function
        # return_api_version doesn't support version 3.0
        self.assertRaises(exceptions.VersionNotFoundForAPIMethod,
                          manager.return_api_version)


class UpdateHeadersTestCase(utils.TestCase):
    def test_api_version_is_null(self):
        headers = {}
        api_versions.update_headers(headers, api_versions.APIVersion())
        self.assertEqual({}, headers)

    def test_api_version_is_major(self):
        headers = {}
        api_versions.update_headers(headers, api_versions.APIVersion("7.0"))
        self.assertEqual({}, headers)

    def test_api_version_is_not_null(self):
        api_version = api_versions.APIVersion("2.3")
        headers = {}
        api_versions.update_headers(headers, api_version)
        self.assertEqual(
            {"OpenStack-API-Version": "volume " + api_version.get_string()},
            headers)


class GetAPIVersionTestCase(utils.TestCase):
    def test_get_available_client_versions(self):
        output = api_versions.get_available_major_versions()
        self.assertNotEqual([], output)

    def test_wrong_format(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.get_api_version, "something_wrong")

    def test_wrong_major_version(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versions.get_api_version, "4")

    @mock.patch("cinderclient.api_versions.get_available_major_versions")
    @mock.patch("cinderclient.api_versions.APIVersion")
    def test_only_major_part_is_presented(self, mock_apiversion,
                                          mock_get_majors):
        mock_get_majors.return_value = [
            str(mock_apiversion.return_value.ver_major)]
        version = 7
        self.assertEqual(mock_apiversion.return_value,
                         api_versions.get_api_version(version))
        mock_apiversion.assert_called_once_with("%s.0" % str(version))

    @mock.patch("cinderclient.api_versions.get_available_major_versions")
    @mock.patch("cinderclient.api_versions.APIVersion")
    def test_major_and_minor_parts_is_presented(self, mock_apiversion,
                                                mock_get_majors):
        version = "2.7"
        mock_get_majors.return_value = [
            str(mock_apiversion.return_value.ver_major)]
        self.assertEqual(mock_apiversion.return_value,
                         api_versions.get_api_version(version))
        mock_apiversion.assert_called_once_with(version)


@ddt.ddt
class DiscoverVersionTestCase(utils.TestCase):
    def setUp(self):
        super(DiscoverVersionTestCase, self).setUp()
        self.orig_max = api_versions.MAX_VERSION
        self.orig_min = api_versions.MIN_VERSION or None
        self.addCleanup(self._clear_fake_version)
        self.fake_client = mock.MagicMock()

    def _clear_fake_version(self):
        api_versions.MAX_VERSION = self.orig_max
        api_versions.MIN_VERSION = self.orig_min

    def _mock_returned_server_version(self, server_version,
                                      server_min_version):
        version_mock = mock.MagicMock(version=server_version,
                                      min_version=server_min_version,
                                      status='CURRENT')
        val = [version_mock]
        if not server_version and not server_min_version:
            val = []
        self.fake_client.services.server_api_version.return_value = val

    @ddt.data(
        # what the data mean:
        # items 1, 2: client min, max
        # items 3, 4: server min, max
        # item 5: user's requested API version
        # item 6: should this raise an exception?
        # item 7: version that should be returned when no exception
        # item 8: what client.services.server_api_version should return
        #         when called by _get_server_version_range in discover_version
        ("3.1", "3.3", "3.4", "3.7", "3.3", True),   # Server too new
        ("3.9", "3.10", "3.0", "3.3", "3.10", True),   # Server too old
        ("3.3", "3.9", "3.7", "3.17", "3.9", False),  # Requested < server
        # downgraded because of server:
        ("3.5", "3.8", "3.0", "3.7", "3.8", False, "3.7"),
        # downgraded because of client:
        ("3.5", "3.8", "3.0", "3.9", "3.9", False, "3.8"),
        # downgraded because of both:
        ("3.5", "3.7", "3.0", "3.8", "3.9", False, "3.7"),
        ("3.5", "3.5", "3.0", "3.5", "3.5", False),  # Server & client same
        ("3.5", "3.5", None, None, "3.5", True, None, []),  # Pre-micro
        ("3.1", "3.11", "3.4", "3.7", "3.7", False),  # Requested in range
        ("3.5", "3.5", "3.0", "3.5", "1.0", True)    # Requested too old
    )
    @ddt.unpack
    def test_microversion(self, client_min, client_max, server_min, server_max,
                          requested_version, exp_range, end_version=None,
                          ret_val=None):
        if ret_val is not None:
            self.fake_client.services.server_api_version.return_value = ret_val
        else:
            self._mock_returned_server_version(server_max, server_min)

        api_versions.MAX_VERSION = client_max
        api_versions.MIN_VERSION = client_min

        if exp_range:
            exc = self.assertRaises(exceptions.UnsupportedVersion,
                                    api_versions.discover_version,
                                    self.fake_client,
                                    api_versions.APIVersion(requested_version))
            if ret_val is not None:
                self.assertIn("Server does not support microversions",
                              str(exc))
            else:
                self.assertIn("range is '%s' to '%s'" %
                              (server_min, server_max), str(exc))
        else:
            discovered_version = api_versions.discover_version(
                self.fake_client,
                api_versions.APIVersion(requested_version))

            version = requested_version
            if end_version is not None:
                version = end_version
            self.assertEqual(version,
                             discovered_version.get_string())
            self.assertTrue(
                self.fake_client.services.server_api_version.called)

    def test_get_highest_version(self):
        self._mock_returned_server_version("3.14", "3.0")
        highest_version = api_versions.get_highest_version(self.fake_client)
        self.assertEqual("3.14", highest_version.get_string())
        self.assertTrue(self.fake_client.services.server_api_version.called)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import api_versions
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes
from cinderclient.v3 import services


cs = fakes.FakeClient(api_version=api_versions.APIVersion('3.0'))


class ServicesTest(utils.TestCase):
    """Tests for v3.0 behavior"""

    def test_list_services(self):
        svs = cs.services.list()
        cs.assert_called('GET', '/os-services')
        self.assertEqual(3, len(svs))
        for service in svs:
            self.assertIsInstance(service, services.Service)
            # Make sure cluster fields from v3.7 are not there
            self.assertFalse(hasattr(service, 'cluster'))
        self._assert_request_id(svs)

    def test_list_services_with_hostname(self):
        svs = cs.services.list(host='host2')
        cs.assert_called('GET', '/os-services?host=host2')
        self.assertEqual(2, len(svs))
        [self.assertIsInstance(s, services.Service) for s in svs]
        [self.assertEqual('host2', s.host) for s in svs]
        self._assert_request_id(svs)

    def test_list_services_with_binary(self):
        svs = cs.services.list(binary='cinder-volume')
        cs.assert_called('GET', '/os-services?binary=cinder-volume')
        self.assertEqual(2, len(svs))
        [self.assertIsInstance(s, services.Service) for s in svs]
        [self.assertEqual('cinder-volume', s.binary) for s in svs]
        self._assert_request_id(svs)

    def test_list_services_with_host_binary(self):
        svs = cs.services.list('host2', 'cinder-volume')
        cs.assert_called('GET', '/os-services?host=host2&binary=cinder-volume')
        self.assertEqual(1, len(svs))
        [self.assertIsInstance(s, services.Service) for s in svs]
        [self.assertEqual('host2', s.host) for s in svs]
        [self.assertEqual('cinder-volume', s.binary) for s in svs]
        self._assert_request_id(svs)

    def test_services_enable(self):
        s = cs.services.enable('host1', 'cinder-volume')
        values = {"host": "host1", 'binary': 'cinder-volume'}
        cs.assert_called('PUT', '/os-services/enable', values)
        self.assertIsInstance(s, services.Service)
        self.assertEqual('enabled', s.status)
        self._assert_request_id(s)

    def test_services_disable(self):
        s = cs.services.disable('host1', 'cinder-volume')
        values = {"host": "host1", 'binary': 'cinder-volume'}
        cs.assert_called('PUT', '/os-services/disable', values)
        self.assertIsInstance(s, services.Service)
        self.assertEqual('disabled', s.status)
        self._assert_request_id(s)

    def test_services_disable_log_reason(self):
        s = cs.services.disable_log_reason(
            'host1', 'cinder-volume', 'disable bad host')
        values = {"host": "host1", 'binary': 'cinder-volume',
                  "disabled_reason": "disable bad host"}
        cs.assert_called('PUT', '/os-services/disable-log-reason', values)
        self.assertIsInstance(s, services.Service)
        self.assertEqual('disabled', s.status)
        self._assert_request_id(s)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Volume snapshot interface (v3 extension)."""

from oslo_utils import strutils

from cinderclient import api_versions
from cinderclient.apiclient import base as common_base
from cinderclient import base

MV_3_66_FORCE_FLAG_ERROR = (
    "Since microversion 3.66 of the Block Storage API,  the 'force' option is "
    "invalid for this request.  For backward compatibility, however, when the "
    "'force' flag is passed with a value evaluating to True, it is silently "
    "ignored.")


class Snapshot(base.Resource):
    """A Snapshot is a point-in-time snapshot of an openstack volume."""

    def __repr__(self):
        return "<Snapshot: %s>" % self.id

    def delete(self, force=False):
        """Delete this snapshot."""
        return self.manager.delete(self, force)

    def update(self, **kwargs):
        """Update the name or description for this snapshot."""
        return self.manager.update(self, **kwargs)

    @property
    def progress(self):
        return self._info.get('os-extended-snapshot-attributes:progress')

    @property
    def project_id(self):
        return self._info.get('os-extended-snapshot-attributes:project_id')

    def reset_state(self, state):
        """Update the snapshot with the provided state."""
        return self.manager.reset_state(self, state)

    def set_metadata(self, metadata):
        """Set metadata of this snapshot."""
        return self.manager.set_metadata(self, metadata)

    def delete_metadata(self, keys):
        """Delete metadata of this snapshot."""
        return self.manager.delete_metadata(self, keys)

    def update_all_metadata(self, metadata):
        """Update_all metadata of this snapshot."""
        return self.manager.update_all_metadata(self, metadata)

    def manage(self, volume_id, ref, name=None, description=None,
               metadata=None):
        """Manage an existing snapshot."""
        self.manager.manage(volume_id=volume_id, ref=ref, name=name,
                            description=description, metadata=metadata)

    def list_manageable(self, host, detailed=True, marker=None, limit=None,
                        offset=None, sort=None, cluster=None):
        return self.manager.list_manageable(host, detailed=detailed,
                                            marker=marker, limit=limit,
                                            offset=offset, sort=sort,
                                            cluster=cluster)

    def unmanage(self, snapshot):
        """Unmanage a snapshot."""
        self.manager.unmanage(snapshot)


class SnapshotManager(base.ManagerWithFind):
    """Manage :class:`Snapshot` resources."""
    resource_class = Snapshot

    @api_versions.wraps("3.0", "3.65")
    def create(self, volume_id, force=False,
               name=None, description=None, metadata=None):

        """Creates a snapshot of the given volume.

        :param volume_id: The ID of the volume to snapshot.
        :param force: If force is True, create a snapshot even if the volume is
        attached to an instance. Default is False.
        :param name: Name of the snapshot
        :param description: Description of the snapshot
        :param metadata: Metadata of the snapshot
        :rtype: :class:`Snapshot`
        """

        if metadata is None:
            snapshot_metadata = {}
        else:
            snapshot_metadata = metadata

        # Bug #1995883: it's possible for the shell to use the user-
        # specified 3.66 do_snapshot_create function, but if the server
        # only supports < 3.66, the client will have been downgraded and
        # will use this function.  In that case, the 'force' parameter will
        # be None, which means that the user didn't specify a value for it,
        # so we set it to the pre-3.66 default value of False.
        #
        # NOTE: we know this isn't a problem for current client consumers
        # because a null value for 'force' has never been allowed by the
        # Block Storage API v3, so there's no reason for anyone to directly
        # call this method passing force=None.
        if force is None:
            force = False

        body = {'snapshot': {'volume_id': volume_id,
                             'force': force,
                             'name': name,
                             'description': description,
                             'metadata': snapshot_metadata}}
        return self._create('/snapshots', body, 'snapshot')

    @api_versions.wraps("3.66")
    def create(self, volume_id, force=None,  # noqa: F811
               name=None, description=None, metadata=None):

        """Creates a snapshot of the given volume.

        :param volume_id: The ID of the volume to snapshot.
        :param force: This is technically not valid after mv 3.66, but the
        API silently accepts force=True for backward compatibility, so this
        function will, too
        :param name: Name of the snapshot
        :param description: Description of the snapshot
        :param metadata: Metadata of the snapshot
        :raises: ValueError if 'force' is not passed with a value that
        evaluates to true
        :rtype: :class:`Snapshot`
        """

        if metadata is None:
            snapshot_metadata = {}
        else:
            snapshot_metadata = metadata

        body = {'snapshot': {'volume_id': volume_id,
                             'name': name,
                             'description': description,
                             'metadata': snapshot_metadata}}
        if force is not None:
            try:
                force = strutils.bool_from_string(force, strict=True)
                if not force:
                    raise ValueError()
            except ValueError:
                raise ValueError(MV_3_66_FORCE_FLAG_ERROR)
        return self._create('/snapshots', body, 'snapshot')

    def get(self, snapshot_id):
        """Shows snapshot details.

        :param snapshot_id: The ID of the snapshot to get.
        :rtype: :class:`Snapshot`
        """
        return self._get("/snapshots/%s" % snapshot_id, "snapshot")

    def list(self, detailed=True, search_opts=None, marker=None, limit=None,
             sort=None):
        """Get a list of all snapshots.

        :rtype: list of :class:`Snapshot`
        """
        resource_type = "snapshots"
        url = self._build_list_url(resource_type, detailed=detailed,
                                   search_opts=search_opts, marker=marker,
                                   limit=limit, sort=sort)
        return self._list(url, resource_type, limit=limit)

    def delete(self, snapshot, force=False):
        """Delete a snapshot.

        :param snapshot: The :class:`Snapshot` to delete.
        :param force: Allow delete in state other than error or available.
        """
        if force:
            return self._action('os-force_delete', snapshot)
        else:
            return self._delete("/snapshots/%s" % base.getid(snapshot))

    def update(self, snapshot, **kwargs):
        """Update the name or description for a snapshot.

        :param snapshot: The :class:`Snapshot` to update.
        """
        if not kwargs:
            return

        body = {"snapshot": kwargs}

        return self._update("/snapshots/%s" % base.getid(snapshot), body)

    def reset_state(self, snapshot, state):
        """Update the specified snapshot with the provided state."""
        return self._action('os-reset_status', snapshot,
                            {'status': state} if state else {})

    def _action(self, action, snapshot, info=None, **kwargs):
        """Perform a snapshot action."""
        body = {action: info}
        self.run_hooks('modify_body_for_action', body, **kwargs)
        url = '/snapshots/%s/action' % base.getid(snapshot)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)

    def update_snapshot_status(self, snapshot, update_dict):
        return self._action('os-update_snapshot_status',
                            base.getid(snapshot), update_dict)

    def set_metadata(self, snapshot, metadata):
        """Update/Set a snapshots metadata.

        :param snapshot: The :class:`Snapshot`.
        :param metadata: A list of keys to be set.
        """
        body = {'metadata': metadata}
        return self._create("/snapshots/%s/metadata" % base.getid(snapshot),
                            body, "metadata")

    def delete_metadata(self, snapshot, keys):
        """Delete specified keys from snapshot metadata.

        :param snapshot: The :class:`Snapshot`.
        :param keys: A list of keys to be removed.
        """
        response_list = []
        snapshot_id = base.getid(snapshot)
        for k in keys:
            resp, body = self._delete("/snapshots/%s/metadata/%s" %
                                      (snapshot_id, k))
            response_list.append(resp)

        return common_base.ListWithMeta([], response_list)

    def update_all_metadata(self, snapshot, metadata):
        """Update_all snapshot metadata.

        :param snapshot: The :class:`Snapshot`.
        :param metadata: A list of keys to be updated.
        """
        body = {'metadata': metadata}
        return self._update("/snapshots/%s/metadata" % base.getid(snapshot),
                            body)

    def manage(self, volume_id, ref, name=None, description=None,
               metadata=None):
        """Manage an existing snapshot."""
        body = {'snapshot': {'volume_id': volume_id,
                             'ref': ref,
                             'name': name,
                             'description': description,
                             'metadata': metadata
                             }
                }
        return self._create('/os-snapshot-manage', body, 'snapshot')

    @api_versions.wraps("3.0")
    def list_manageable(self, host, detailed=True, marker=None,
                        limit=None, offset=None, sort=None):
        url = self._build_list_url("os-snapshot-manage", detailed=detailed,
                                   search_opts={'host': host}, marker=marker,
                                   limit=limit, offset=offset, sort=sort)
        return self._list(url, "manageable-snapshots")

    @api_versions.wraps('3.8')
    def list_manageable(self, host, detailed=True, marker=None,  # noqa: F811
                        limit=None, offset=None, sort=None, cluster=None):
        search_opts = {'cluster': cluster} if cluster else {'host': host}
        url = self._build_list_url("manageable_snapshots", detailed=detailed,
                                   search_opts=search_opts, marker=marker,
                                   limit=limit, offset=offset, sort=sort)
        return self._list(url, "manageable-snapshots")

    def unmanage(self, snapshot):
        """Unmanage a snapshot."""
        return self._action('os-unmanage', snapshot, None)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
service interface
"""

from cinderclient import api_versions
from cinderclient import base


class Service(base.Resource):

    def __repr__(self):
        return "<Service: binary=%s host=%s>" % (self.binary, self.host)


class LogLevel(base.Resource):
    def __repr__(self):
        return '<LogLevel: binary=%s host=%s prefix=%s level=%s>' % (
            self.binary, self.host, self.prefix, self.level)


class ServiceManagerBase(base.ManagerWithFind):
    resource_class = Service

    def list(self, host=None, binary=None):
        """
        Describes service list for host.

        :param host: destination host name.
        :param binary: service binary.
        """
        url = "/os-services"
        filters = []
        if host:
            filters.append("host=%s" % host)
        if binary:
            filters.append("binary=%s" % binary)
        if filters:
            url = "%s?%s" % (url, "&".join(filters))
        return self._list(url, "services")

    def enable(self, host, binary):
        """Enable the service specified by hostname and binary."""
        body = {"host": host, "binary": binary}
        result = self._update("/os-services/enable", body)
        return self.resource_class(self, result, resp=result.request_ids)

    def disable(self, host, binary):
        """Disable the service specified by hostname and binary."""
        body = {"host": host, "binary": binary}
        result = self._update("/os-services/disable", body)
        return self.resource_class(self, result, resp=result.request_ids)

    def disable_log_reason(self, host, binary, reason):
        """Disable the service with reason."""
        body = {"host": host, "binary": binary, "disabled_reason": reason}
        result = self._update("/os-services/disable-log-reason", body)
        return self.resource_class(self, result, resp=result.request_ids)

    def freeze_host(self, host):
        """Freeze the service specified by hostname."""
        body = {"host": host}
        return self._update("/os-services/freeze", body)

    def thaw_host(self, host):
        """Thaw the service specified by hostname."""
        body = {"host": host}
        return self._update("/os-services/thaw", body)

    def failover_host(self, host, backend_id):
        """Failover a replicated backend by hostname."""
        body = {"host": host, "backend_id": backend_id}
        return self._update("/os-services/failover_host", body)


class ServiceManager(ServiceManagerBase):
    @api_versions.wraps("3.0")
    def server_api_version(self):
        """Returns the API Version supported by the server.

        :return: Returns response obj for a server that supports microversions.
                 Returns an empty list for Liberty and prior Cinder servers.
        """

        try:
            return self._get_with_base_url("", response_key='versions')
        except LookupError:
            return []

    @api_versions.wraps("3.32")
    def set_log_levels(self, level, binary, server, prefix):
        """Set log level for services."""
        body = {'level': level, 'binary': binary, 'server': server,
                'prefix': prefix}
        return self._update("/os-services/set-log", body)

    @api_versions.wraps("3.32")
    def get_log_levels(self, binary, server, prefix):
        """Get log levels for services."""
        body = {'binary': binary, 'server': server, 'prefix': prefix}
        response = self._update("/os-services/get-log", body)

        log_levels = []
        for entry in response['log_levels']:
            entry_levels = sorted(entry['levels'].items(), key=lambda x: x[0])
            for prefix, level in entry_levels:
                log_dict = {'binary': entry['binary'], 'host': entry['host'],
                            'prefix': prefix, 'level': level}
                log_levels.append(LogLevel(self, log_dict, loaded=True))
        return log_levels



# Copyright (C) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Group interface (v3 extension)."""
from cinderclient import api_versions
from cinderclient.apiclient import base as common_base
from cinderclient import base
from cinderclient import utils


class Group(base.Resource):
    """A Group of volumes."""
    def __repr__(self):
        return "<Group: %s>" % self.id

    def delete(self, delete_volumes=False):
        """Delete this group."""
        return self.manager.delete(self, delete_volumes)

    def update(self, **kwargs):
        """Update the name or description for this group."""
        return self.manager.update(self, **kwargs)

    def reset_state(self, state):
        """Reset the group's state with specified one"""
        return self.manager.reset_state(self, state)

    def enable_replication(self):
        """Enables replication for this group."""
        return self.manager.enable_replication(self)

    def disable_replication(self):
        """Disables replication for this group."""
        return self.manager.disable_replication(self)

    def failover_replication(self, allow_attached_volume=False,
                             secondary_backend_id=None):
        """Fails over replication for this group."""
        return self.manager.failover_replication(self,
            allow_attached_volume,
            secondary_backend_id)

    def list_replication_targets(self):
        """Lists replication targets for this group."""
        return self.manager.list_replication_targets(self)


class GroupManager(base.ManagerWithFind):
    """Manage :class:`Group` resources."""
    resource_class = Group

    @api_versions.wraps('3.13')
    def create(self, group_type, volume_types, name=None,
               description=None, user_id=None,
               project_id=None, availability_zone=None):
        """Creates a group.

        :param group_type: Type of the Group
        :param volume_types: Types of volume
        :param name: Name of the Group
        :param description: Description of the Group
        :param user_id: User id derived from context
        :param project_id: Project id derived from context
        :param availability_zone: Availability Zone to use
        :rtype: :class:`Group`
        """
        body = {'group': {'name': name,
                          'description': description,
                          'group_type': group_type,
                          'volume_types': volume_types.split(','),
                          'availability_zone': availability_zone,
                          }}

        return self._create('/groups', body, 'group')

    @api_versions.wraps('3.20')
    def reset_state(self, group, state):
        """Update the provided group with the provided state.

        :param group: The :class:`Group` to set the state.
        :param state: The state of the group to be set.
        """
        body = {'status': state} if state else {}
        return self._action('reset_status', group, body)

    @api_versions.wraps('3.14')
    def create_from_src(self, group_snapshot_id, source_group_id,
                        name=None, description=None, user_id=None,
                        project_id=None):
        """Creates a group from a group snapshot or a source group.

        :param group_snapshot_id: UUID of a GroupSnapshot
        :param source_group_id: UUID of a source Group
        :param name: Name of the Group
        :param description: Description of the Group
        :param user_id: User id derived from context
        :param project_id: Project id derived from context
        :rtype: A dictionary containing Group metadata
        """

        # NOTE(wanghao): According the API schema in cinder side, client
        # should NOT specify the group_snapshot_id and source_group_id at
        # same time, even one of them is None.
        if group_snapshot_id:
            create_key = 'group_snapshot_id'
            create_value = group_snapshot_id
        elif source_group_id:
            create_key = 'source_group_id'
            create_value = source_group_id

        body = {'create-from-src': {'name': name,
                                    'description': description,
                                    create_key: create_value}}

        self.run_hooks('modify_body_for_action', body,
                       'create-from-src')
        resp, body = self.api.client.post(
            "/groups/action", body=body)
        return common_base.DictWithMeta(body['group'], resp)

    @api_versions.wraps('3.13')
    def get(self, group_id, **kwargs):
        """Get a group.

        :param group_id: The ID of the group to get.
        :rtype: :class:`Group`
        """
        query_params = kwargs
        query_string = utils.build_query_param(query_params, sort=True)

        return self._get("/groups/%s" % group_id + query_string,
                         "group")

    @api_versions.wraps('3.13')
    def list(self, detailed=True, search_opts=None, list_volume=False):
        """Lists all groups.

        :rtype: list of :class:`Group`
        """
        if list_volume:
            if not search_opts:
                search_opts = {}
            search_opts['list_volume'] = True
        query_string = utils.build_query_param(search_opts, sort=True)

        detail = ""
        if detailed:
            detail = "/detail"

        return self._list("/groups%s%s" % (detail, query_string),
                          "groups")

    @api_versions.wraps('3.13')
    def delete(self, group, delete_volumes=False):
        """Delete a group.

        :param group: the :class:`Group` to delete.
        :param delete_volumes: delete volumes in the group.
        """
        body = {'delete': {'delete-volumes': delete_volumes}}
        self.run_hooks('modify_body_for_action', body, 'group')
        url = '/groups/%s/action' % base.getid(group)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)

    @api_versions.wraps('3.13')
    def update(self, group, **kwargs):
        """Update the name or description for a group.

        :param Group: The :class:`Group` to update.
        """
        if not kwargs:
            return

        body = {"group": kwargs}

        return self._update("/groups/%s" %
                            base.getid(group), body)

    def _action(self, action, group, info=None, **kwargs):
        """Perform a group "action."

        :param action: an action to be performed on the group
        :param group: a group to perform the action on
        :param info: details of the action
        :param **kwargs: other parameters
        """

        body = {action: info}
        self.run_hooks('modify_body_for_action', body, **kwargs)
        url = '/groups/%s/action' % base.getid(group)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)

    @api_versions.wraps('3.38')
    def enable_replication(self, group):
        """Enables replication for a group.

        :param group: the :class:`Group` to enable replication.
        """
        body = {'enable_replication': {}}
        self.run_hooks('modify_body_for_action', body, 'group')
        url = '/groups/%s/action' % base.getid(group)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)

    @api_versions.wraps('3.38')
    def disable_replication(self, group):
        """disables replication for a group.

        :param group: the :class:`Group` to disable replication.
        """
        body = {'disable_replication': {}}
        self.run_hooks('modify_body_for_action', body, 'group')
        url = '/groups/%s/action' % base.getid(group)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)

    @api_versions.wraps('3.38')
    def failover_replication(self, group, allow_attached_volume=False,
                             secondary_backend_id=None):
        """fails over replication for a group.

        :param group: the :class:`Group` to failover.
        :param allow attached volumes: allow attached volumes in the group.
        :param secondary_backend_id: secondary backend id.
        """
        body = {
            'failover_replication': {
                'allow_attached_volume': allow_attached_volume,
                'secondary_backend_id': secondary_backend_id
            }
        }
        self.run_hooks('modify_body_for_action', body, 'group')
        url = '/groups/%s/action' % base.getid(group)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)

    @api_versions.wraps('3.38')
    def list_replication_targets(self, group):
        """List replication targets for a group.

        :param group: the :class:`Group` to list replication targets.
        """
        body = {'list_replication_targets': {}}
        self.run_hooks('modify_body_for_action', body, 'group')
        url = '/groups/%s/action' % base.getid(group)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)



# Copyright (c) 2016 EMC Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Group Type interface."""

from urllib import parse

from cinderclient import api_versions
from cinderclient import base


class GroupType(base.Resource):
    """A Group Type is the type of group to be created."""
    def __repr__(self):
        return "<GroupType: %s>" % self.name

    @property
    def is_public(self):
        """
        Provide a user-friendly accessor to is_public
        """
        return self._info.get("is_public",
                              self._info.get("is_public", 'N/A'))

    @api_versions.wraps("3.11")
    def get_keys(self):
        """Get group specs from a group type.

        :param type: The :class:`GroupType` to get specs from
        """
        _resp, body = self.manager.api.client.get(
            "/group_types/%s/group_specs" %
            base.getid(self))
        return body["group_specs"]

    @api_versions.wraps("3.11")
    def set_keys(self, metadata):
        """Set group specs on a group type.

        :param type : The :class:`GroupType` to set spec on
        :param metadata: A dict of key/value pairs to be set
        """
        body = {'group_specs': metadata}
        return self.manager._create(
            "/group_types/%s/group_specs" % base.getid(self),
            body,
            "group_specs",
            return_raw=True)

    @api_versions.wraps("3.11")
    def unset_keys(self, keys):
        """Unset specs on a group type.

        :param type_id: The :class:`GroupType` to unset spec on
        :param keys: A list of keys to be unset
        """

        for k in keys:
            resp = self.manager._delete(
                "/group_types/%s/group_specs/%s" % (
                base.getid(self), k))
            if resp:
                return resp


class GroupTypeManager(base.ManagerWithFind):
    """Manage :class:`GroupType` resources."""
    resource_class = GroupType

    @api_versions.wraps("3.11")
    def list(self, search_opts=None, is_public=None):
        """Lists all group types.

        :rtype: list of :class:`GroupType`.
        """
        if not search_opts:
            search_opts = dict()

        query_string = ''
        if 'is_public' not in search_opts:
            search_opts['is_public'] = is_public

        query_string = "?%s" % parse.urlencode(search_opts)
        return self._list("/group_types%s" % (query_string), "group_types")

    @api_versions.wraps("3.11")
    def get(self, group_type):
        """Get a specific group type.

        :param group_type: The ID of the :class:`GroupType` to get.
        :rtype: :class:`GroupType`
        """
        return self._get("/group_types/%s" % base.getid(group_type),
                         "group_type")

    @api_versions.wraps("3.11")
    def default(self):
        """Get the default group type.

        :rtype: :class:`GroupType`
        """
        return self._get("/group_types/default", "group_type")

    @api_versions.wraps("3.11")
    def delete(self, group_type):
        """Deletes a specific group_type.

        :param group_type: The name or ID of the :class:`GroupType` to get.
        """
        return self._delete("/group_types/%s" % base.getid(group_type))

    @api_versions.wraps("3.11")
    def create(self, name, description=None, is_public=True):
        """Creates a group type.

        :param name: Descriptive name of the group type
        :param description: Description of the group type
        :param is_public: Group type visibility
        :rtype: :class:`GroupType`
        """

        body = {
            "group_type": {
                "name": name,
                "description": description,
                "is_public": is_public,
            }
        }

        return self._create("/group_types", body, "group_type")

    @api_versions.wraps("3.11")
    def update(self, group_type, name=None, description=None, is_public=None):
        """Update the name and/or description for a group type.

        :param group_type: The ID of the :class:`GroupType` to update.
        :param name: Descriptive name of the group type.
        :param description: Description of the group type.
        :rtype: :class:`GroupType`
        """

        body = {
            "group_type": {
                "name": name,
                "description": description
            }
        }
        if is_public is not None:
            body["group_type"]["is_public"] = is_public

        return self._update("/group_types/%s" % base.getid(group_type),
                            body, response_key="group_type")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import fixture

from cinderclient.tests.unit.fixture_data import base
from cinderclient.v3 import client as v3client


class Base(base.Fixture):

    def __init__(self, *args, **kwargs):
        super(Base, self).__init__(*args, **kwargs)

        self.token = fixture.V2Token()
        self.token.set_scope()

    def setUp(self):
        super(Base, self).setUp()

        auth_url = '%s/tokens' % self.identity_url
        self.requests.register_uri('POST', auth_url,
                                   json=self.token,
                                   headers=self.json_headers)


class V3(Base):

    def __init__(self, *args, **kwargs):
        super(V3, self).__init__(*args, **kwargs)

        svc = self.token.add_service('volumev3')
        svc.add_endpoint(self.volume_url)

    def new_client(self):
        return v3client.Client(username='xx',
                               api_key='xx',
                               project_id='xx',
                               auth_url=self.identity_url)



# Copyright 2016 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import api_versions
from cinderclient import utils


@api_versions.wraps("3.0", "3.1")
def do_fake_action():
    """help message

    This will not show up in help message
    """
    return "fake_action 3.0 to 3.1"


@api_versions.wraps("3.2", "3.3")
def do_fake_action():  # noqa: F811
    return "fake_action 3.2 to 3.3"


@api_versions.wraps("3.6")
@utils.arg(
    '--foo',
    start_version='3.7')
def do_another_fake_action():
    return "another_fake_action"


@utils.arg(
    '--foo',
    start_version='3.1',
    end_version='3.2')
@utils.arg(
    '--bar',
    help='bar help',
    start_version='3.3',
    end_version='3.4')
def do_fake_action2():
    return "fake_action2"


@utils.arg(
    '--foo',
    help='first foo',
    start_version='3.6',
    end_version='3.7')
@utils.arg(
    '--foo',
    help='second foo',
    start_version='3.8')
def do_fake_action3():
    return "fake_action3"



# Copyright (c) 2013 OpenStack Foundation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import extension
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes
from cinderclient.v3.contrib import list_extensions

extensions = [
    extension.Extension(list_extensions.__name__.split(".")[-1],
                        list_extensions),
]
cs = fakes.FakeClient(extensions=extensions)


class ListExtensionsTests(utils.TestCase):
    def test_list_extensions(self):
        all_exts = cs.list_extensions.show_all()
        cs.assert_called('GET', '/extensions')
        self.assertGreater(len(all_exts), 0)
        for r in all_exts:
            self.assertGreater(len(r.summary), 0)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from unittest import mock

import ddt
import fixtures
from keystoneauth1 import adapter
from keystoneauth1 import exceptions as keystone_exception
from oslo_serialization import jsonutils

from cinderclient import api_versions
import cinderclient.client
from cinderclient import exceptions
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes


@ddt.ddt
class ClientTest(utils.TestCase):

    def test_get_client_class_v2(self):
        self.assertRaises(cinderclient.exceptions.UnsupportedVersion,
                          cinderclient.client.get_client_class,
                          '2')

    def test_get_client_class_unknown(self):
        self.assertRaises(cinderclient.exceptions.UnsupportedVersion,
                          cinderclient.client.get_client_class, '0')

    @mock.patch.object(cinderclient.client.HTTPClient, '__init__')
    @mock.patch('cinderclient.client.SessionClient')
    def test_construct_http_client_endpoint_url(
            self, session_mock, httpclient_mock):
        os_endpoint = 'http://example.com/'
        httpclient_mock.return_value = None
        cinderclient.client._construct_http_client(
            os_endpoint=os_endpoint)
        self.assertTrue(httpclient_mock.called)
        self.assertEqual(os_endpoint,
                         httpclient_mock.call_args[1].get('os_endpoint'))
        session_mock.assert_not_called()

    def test_log_req(self):
        self.logger = self.useFixture(
            fixtures.FakeLogger(
                format="%(message)s",
                level=logging.DEBUG,
                nuke_handlers=True
            )
        )

        kwargs = {
            'headers': {"X-Foo": "bar"},
            'data': ('{"auth": {"tenantName": "fakeService",'
                     ' "passwordCredentials": {"username": "fakeUser",'
                     ' "password": "fakePassword"}}}')
        }

        cs = cinderclient.client.HTTPClient("user", None, None,
                                            "http://127.0.0.1:5000")
        cs.http_log_debug = True
        cs.http_log_req('PUT', kwargs)

        output = self.logger.output.split('\n')

        self.assertNotIn("fakePassword", output[1])
        self.assertIn("fakeUser", output[1])

    def test_versions(self):
        v2_url = 'http://fakeurl/v2/tenants'
        v3_url = 'http://fakeurl/v3/tenants'
        unknown_url = 'http://fakeurl/v9/tenants'

        self.assertRaises(cinderclient.exceptions.UnsupportedVersion,
                          cinderclient.client.get_volume_api_from_url,
                          v2_url)
        self.assertEqual('3',
                         cinderclient.client.get_volume_api_from_url(v3_url))
        self.assertRaises(cinderclient.exceptions.UnsupportedVersion,
                          cinderclient.client.get_volume_api_from_url,
                          unknown_url)

    @mock.patch('cinderclient.client.SessionClient.get_endpoint')
    @ddt.data(
        ('http://192.168.1.1:8776/v2', 'http://192.168.1.1:8776/'),
        ('http://192.168.1.1:8776/v3/e5526285ebd741b1819393f772f11fc3',
         'http://192.168.1.1:8776/'),
        ('https://192.168.1.1:8080/volumes/v3/'
         'e5526285ebd741b1819393f772f11fc3',
         'https://192.168.1.1:8080/volumes/'),
        ('http://192.168.1.1/volumes/v3/e5526285ebd741b1819393f772f11fc3',
         'http://192.168.1.1/volumes/'),
        ('https://volume.example.com/', 'https://volume.example.com/'))
    @ddt.unpack
    def test_get_base_url(self, url, expected_base, mock_get_endpoint):
        mock_get_endpoint.return_value = url
        cs = cinderclient.client.SessionClient(self, api_version='3.0')
        self.assertEqual(expected_base, cs._get_base_url())

    @mock.patch.object(adapter.LegacyJsonAdapter, '_request')
    @mock.patch.object(exceptions, 'from_response')
    def test_sessionclient_request_method(
            self, mock_from_resp, mock_request):
        kwargs = {
            "body": {
                "volume": {
                    "status": "creating",
                    "imageRef": "username",
                    "attach_status": "detached"
                },
                "authenticated": "True"
            }
        }

        resp = {
            "text": {
                "volume": {
                    "status": "creating",
                    "id": "431253c0-e203-4da2-88df-60c756942aaf",
                    "size": 1
                }
            },
            "code": 202
        }

        request_id = "req-f551871a-4950-4225-9b2c-29a14c8f075e"
        mock_response = utils.TestResponse({
            "status_code": 202,
            "text": json.dumps(resp).encode("latin-1"),
            "headers": {"x-openstack-request-id": request_id},
        })

        # 'request' method of Adaptor will return 202 response
        mock_request.return_value = mock_response
        session_client = cinderclient.client.SessionClient(session=mock.Mock())
        response, body = session_client.request(mock.sentinel.url,
                                                'POST', **kwargs)
        self.assertIsNotNone(session_client._logger)

        # In this case, from_response method will not get called
        # because response status_code is < 400
        self.assertEqual(202, response.status_code)
        self.assertFalse(mock_from_resp.called)

    @mock.patch.object(adapter.LegacyJsonAdapter, '_request')
    def test_sessionclient_request_method_raises_badrequest(
            self, mock_request):
        kwargs = {
            "body": {
                "volume": {
                    "status": "creating",
                    "imageRef": "username",
                    "attach_status": "detached"
                },
                "authenticated": "True"
            }
        }

        resp = {
            "badRequest": {
                "message": "Invalid image identifier or unable to access "
                           "requested image.",
                "code": 400
            }
        }

        mock_response = utils.TestResponse({
            "status_code": 400,
            "text": json.dumps(resp).encode("latin-1"),
        })

        # 'request' method of Adaptor will return 400 response
        mock_request.return_value = mock_response
        session_client = cinderclient.client.SessionClient(
            session=mock.Mock())

        # 'from_response' method will raise BadRequest because
        # resp.status_code is 400
        self.assertRaises(exceptions.BadRequest, session_client.request,
                          mock.sentinel.url, 'POST', **kwargs)
        self.assertIsNotNone(session_client._logger)

    @mock.patch.object(adapter.LegacyJsonAdapter, '_request')
    def test_sessionclient_request_method_raises_overlimit(
            self, mock_request):
        resp = {
            "overLimitFault": {
                "message": "This request was rate-limited.",
                "code": 413
            }
        }

        mock_response = utils.TestResponse({
            "status_code": 413,
            "text": json.dumps(resp).encode("latin-1"),
        })

        # 'request' method of Adaptor will return 413 response
        mock_request.return_value = mock_response
        session_client = cinderclient.client.SessionClient(
            session=mock.Mock())

        self.assertRaises(exceptions.OverLimit, session_client.request,
                          mock.sentinel.url, 'GET')
        self.assertIsNotNone(session_client._logger)

    @mock.patch.object(exceptions, 'from_response')
    def test_keystone_request_raises_auth_failure_exception(
            self, mock_from_resp):

        kwargs = {
            "body": {
                "volume": {
                    "status": "creating",
                    "imageRef": "username",
                    "attach_status": "detached"
                },
                "authenticated": "True"
            }
        }

        with mock.patch.object(adapter.LegacyJsonAdapter, '_request',
                               side_effect=
                               keystone_exception.AuthorizationFailure()):
            session_client = cinderclient.client.SessionClient(
                session=mock.Mock())
            self.assertRaises(keystone_exception.AuthorizationFailure,
                              session_client.request,
                              mock.sentinel.url, 'POST', **kwargs)

        # As keystonesession.request method will raise
        # AuthorizationFailure exception, check exceptions.from_response
        # is not getting called.
        self.assertFalse(mock_from_resp.called)

    @mock.patch('keystoneauth1.adapter.LegacyJsonAdapter.request',
                return_value=(mock.Mock(), mock.Mock()))
    @ddt.data(True, False, None)
    def test_http_log_debug_request(self, http_log_debug, mock_request):
        args_req = (mock.sentinel.url, mock.sentinel.OP)
        kwargs_req = {'raise_exc': False}
        kwargs_expect = {'authenticated': False}
        kwargs_expect.update(kwargs_req)

        kwargs = {'api_version': '3.0'}
        if isinstance(http_log_debug, bool):
            kwargs['http_log_debug'] = http_log_debug
            if http_log_debug:
                kwargs_expect['logger'] = mock.ANY

        cs = cinderclient.client.SessionClient(self, **kwargs)

        res = cs.request(*args_req, **kwargs_req)

        mock_request.assert_called_once_with(*args_req, **kwargs_expect)
        self.assertEqual(mock_request.return_value, res)


class ClientTestSensitiveInfo(utils.TestCase):
    def test_req_does_not_log_sensitive_info(self):
        self.logger = self.useFixture(
            fixtures.FakeLogger(
                format="%(message)s",
                level=logging.DEBUG,
                nuke_handlers=True
            )
        )

        secret_auth_token = "MY_SECRET_AUTH_TOKEN"
        kwargs = {
            'headers': {"X-Auth-Token": secret_auth_token},
            'data': ('{"auth": {"tenantName": "fakeService",'
                     ' "passwordCredentials": {"username": "fakeUser",'
                     ' "password": "fakePassword"}}}')
        }

        cs = cinderclient.client.HTTPClient("user", None, None,
                                            "http://127.0.0.1:5000")
        cs.http_log_debug = True
        cs.http_log_req('PUT', kwargs)

        output = self.logger.output.split('\n')
        self.assertNotIn(secret_auth_token, output[1])

    def test_resp_does_not_log_sensitive_info(self):
        self.logger = self.useFixture(
            fixtures.FakeLogger(
                format="%(message)s",
                level=logging.DEBUG,
                nuke_handlers=True
            )
        )
        cs = cinderclient.client.HTTPClient("user", None, None,
                                            "http://127.0.0.1:5000")
        resp = mock.Mock()
        resp.status_code = 200
        resp.headers = {
            'x-compute-request-id': 'req-f551871a-4950-4225-9b2c-29a14c8f075e'
        }
        auth_password = "kk4qD6CpKFLyz9JD"
        body = {
            "connection_info": {
                "driver_volume_type": "iscsi",
                "data": {
                    "auth_password": auth_password,
                    "target_discovered": False,
                    "encrypted": False,
                    "qos_specs": None,
                    "target_iqn": ("iqn.2010-10.org.openstack:volume-"
                                   "a2f33dcc-1bb7-45ba-b8fc-5b38179120f8"),
                    "target_portal": "10.0.100.186:3260",
                    "volume_id": "a2f33dcc-1bb7-45ba-b8fc-5b38179120f8",
                    "target_lun": 1,
                    "access_mode": "rw",
                    "auth_username": "s4BfSfZ67Bo2mnpuFWY8",
                    "auth_method": "CHAP"
                }
            }
        }
        resp.text = jsonutils.dumps(body)
        cs.http_log_debug = True
        cs.http_log_resp(resp)

        output = self.logger.output.split('\n')
        self.assertIn('***', output[1], output)
        self.assertNotIn(auth_password, output[1], output)


@ddt.ddt
class GetAPIVersionTestCase(utils.TestCase):

    @mock.patch('cinderclient.client.requests.get')
    def test_get_server_version_v2(self, mock_request):
        # Why are we testing this? Because we can!

        mock_response = utils.TestResponse({
            "status_code": 200,
            "text": json.dumps(fakes.fake_request_get_no_v3())
        })

        mock_request.return_value = mock_response

        url = "http://192.168.122.127:8776/v2/e5526285ebd741b1819393f772f11fc3"

        min_version, max_version = cinderclient.client.get_server_version(url)

        self.assertEqual(api_versions.APIVersion('2.0'), min_version)
        self.assertEqual(api_versions.APIVersion('2.0'), max_version)

    @mock.patch('cinderclient.client.requests.get')
    @ddt.data(
        'http://192.168.122.127:8776/v3/e5526285ebd741b1819393f772f11fc3',
        'https://192.168.122.127:8776/v3/e55285ebd741b1819393f772f11fc3',
        'http://192.168.122.127/volumesv3/e5526285ebd741b1819393f772f11fc3'
        )
    def test_get_server_version(self, url, mock_request):
        mock_response = utils.TestResponse({
            "status_code": 200,
            "text": json.dumps(fakes.fake_request_get())
        })

        mock_request.return_value = mock_response

        min_version, max_version = cinderclient.client.get_server_version(url)
        self.assertEqual(min_version, api_versions.APIVersion('3.0'))
        self.assertEqual(max_version, api_versions.APIVersion('3.16'))

    @mock.patch('cinderclient.client.requests.get')
    def test_get_server_version_insecure(self, mock_request):
        mock_response = utils.TestResponse({
            "status_code": 200,
            "text": json.dumps(fakes.fake_request_get_no_v3())
        })

        mock_request.return_value = mock_response

        url = (
            "https://192.168.122.127:8776/v3/e5526285ebd741b1819393f772f11fc3")
        expected_url = "https://192.168.122.127:8776/"

        cinderclient.client.get_server_version(url, True)

        mock_request.assert_called_once_with(expected_url,
                                             verify=False,
                                             cert=None)

    @mock.patch('cinderclient.client.requests.get')
    def test_get_server_version_cacert(self, mock_request):
        mock_response = utils.TestResponse({
            "status_code": 200,
            "text": json.dumps(fakes.fake_request_get_no_v3())
        })

        mock_request.return_value = mock_response

        url = (
            "https://192.168.122.127:8776/v3/e5526285ebd741b1819393f772f11fc3")
        expected_url = "https://192.168.122.127:8776/"

        cacert = '/path/to/cert'
        cinderclient.client.get_server_version(url, cacert=cacert)

        mock_request.assert_called_once_with(expected_url,
                                             verify=cacert,
                                             cert=None)

    @mock.patch('cinderclient.client.requests.get')
    def test_get_server_version_cert(self, mock_request):
        mock_response = utils.TestResponse({
            "status_code": 200,
            "text": json.dumps(fakes.fake_request_get_no_v3())
        })

        mock_request.return_value = mock_response

        url = (
            "https://192.168.122.127:8776/v3/e5526285ebd741b1819393f772f11fc3")
        expected_url = "https://192.168.122.127:8776/"

        client_cert = '/path/to/cert'
        cinderclient.client.get_server_version(url, cert=client_cert)

        mock_request.assert_called_once_with(expected_url,
                                             verify=True,
                                             cert=client_cert)

    @mock.patch('cinderclient.client.requests.get')
    @ddt.data('3.12', '3.40')
    def test_get_highest_client_server_version(self, version, mock_request):

        mock_response = utils.TestResponse({
            "status_code": 200,
            "text": json.dumps(fakes.fake_request_get())
        })

        mock_request.return_value = mock_response

        url = "http://192.168.122.127:8776/v3/e5526285ebd741b1819393f772f11fc3"

        with mock.patch.object(api_versions, 'MAX_VERSION', version):
            highest = (
                cinderclient.client.get_highest_client_server_version(url))
        expected = version if version == '3.12' else '3.16'
        self.assertEqual(expected, highest)

    @mock.patch('cinderclient.client.requests.get')
    def test_get_highest_client_server_version_negative(self,
                                                        mock_request):

        mock_response = utils.TestResponse({
            "status_code": 200,
            "text": json.dumps(fakes.fake_request_get_no_v3())
        })

        mock_request.return_value = mock_response

        url = "http://192.168.122.127:8776/v3/e5526285ebd741b1819393f772f11fc3"

        self.assertRaises(exceptions.UnsupportedVersion,
                          cinderclient.client.
                          get_highest_client_server_version,
                          url)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from cinderclient.tests.functional import base


class CinderSnapshotTests(base.ClientTestBase):
    """Check of cinder snapshot commands."""
    def setUp(self):
        super(CinderSnapshotTests, self).setUp()
        self.volume = self.object_create('volume', params='1')

    def test_snapshot_create_description(self):
        """Test steps:

        1) create volume in Setup()
        2) create snapshot with description
        3) check that snapshot has right description
        """
        description = 'test_description'
        snapshot = self.object_create('snapshot',
                                      params='--description {0} {1}'.
                                      format(description, self.volume['id']))
        self.assertEqual(description, snapshot['description'])
        self.object_delete('snapshot', snapshot['id'])
        self.check_object_deleted('snapshot', snapshot['id'])

    def test_snapshot_create_metadata(self):
        """Test steps:

        1) create volume in Setup()
        2) create snapshot with metadata
        3) check that metadata complies entered
        """
        snapshot = self.object_create(
            'snapshot',
            params='--metadata test_metadata=test_date {0}'.format(
                self.volume['id']))
        self.assertEqual(str({'test_metadata': 'test_date'}),
                         snapshot['metadata'])
        self.object_delete('snapshot', snapshot['id'])
        self.check_object_deleted('snapshot', snapshot['id'])



# Copyright (C) 2016 EMC Corporation.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from cinderclient import api_versions
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes

cs = fakes.FakeClient(api_versions.APIVersion('3.13'))


@ddt.ddt
class GroupsTest(utils.TestCase):

    def test_delete_group(self):
        expected = {'delete': {'delete-volumes': True}}
        v = cs.groups.list()[0]
        grp = v.delete(delete_volumes=True)
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.delete('1234', delete_volumes=True)
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.delete(v, delete_volumes=True)
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)

    def test_create_group(self):
        grp = cs.groups.create('my_group_type', 'type1,type2', name='group')
        cs.assert_called('POST', '/groups')
        self._assert_request_id(grp)

    def test_create_group_with_volume_types(self):
        grp = cs.groups.create('my_group_type', 'type1,type2', name='group')
        expected = {'group': {'description': None,
                              'availability_zone': None,
                              'name': 'group',
                              'group_type': 'my_group_type',
                              'volume_types': ['type1', 'type2']}}
        cs.assert_called('POST', '/groups', body=expected)
        self._assert_request_id(grp)

    @ddt.data(
        {'name': 'group2', 'desc': None, 'add': None, 'remove': None},
        {'name': None, 'desc': 'group2 desc', 'add': None, 'remove': None},
        {'name': None, 'desc': None, 'add': 'uuid1,uuid2', 'remove': None},
        {'name': None, 'desc': None, 'add': None, 'remove': 'uuid3,uuid4'},
    )
    @ddt.unpack
    def test_update_group_name(self, name, desc, add, remove):
        v = cs.groups.list()[0]
        expected = {'group': {'name': name, 'description': desc,
                              'add_volumes': add, 'remove_volumes': remove}}
        grp = v.update(name=name, description=desc,
                       add_volumes=add, remove_volumes=remove)
        cs.assert_called('PUT', '/groups/1234', body=expected)
        self._assert_request_id(grp)
        grp = cs.groups.update('1234', name=name, description=desc,
                               add_volumes=add, remove_volumes=remove)
        cs.assert_called('PUT', '/groups/1234', body=expected)
        self._assert_request_id(grp)
        grp = cs.groups.update(v, name=name, description=desc,
                               add_volumes=add, remove_volumes=remove)
        cs.assert_called('PUT', '/groups/1234', body=expected)
        self._assert_request_id(grp)

    def test_update_group_none(self):
        self.assertIsNone(cs.groups.update('1234'))

    def test_update_group_no_props(self):
        cs.groups.update('1234')

    def test_list_group(self):
        lst = cs.groups.list()
        cs.assert_called('GET', '/groups/detail')
        self._assert_request_id(lst)

    def test_list_group_detailed_false(self):
        lst = cs.groups.list(detailed=False)
        cs.assert_called('GET', '/groups')
        self._assert_request_id(lst)

    def test_list_group_with_search_opts(self):
        lst = cs.groups.list(search_opts={'foo': 'bar'})
        cs.assert_called('GET', '/groups/detail?foo=bar')
        self._assert_request_id(lst)

    def test_list_group_with_volume(self):
        lst = cs.groups.list(list_volume=True)
        cs.assert_called('GET', '/groups/detail?list_volume=True')
        self._assert_request_id(lst)

    def test_list_group_with_empty_search_opt(self):
        lst = cs.groups.list(
            search_opts={'foo': 'bar', 'abc': None}
        )
        cs.assert_called('GET', '/groups/detail?foo=bar')
        self._assert_request_id(lst)

    def test_get_group(self):
        group_id = '1234'
        grp = cs.groups.get(group_id)
        cs.assert_called('GET', '/groups/%s' % group_id)
        self._assert_request_id(grp)

    def test_get_group_with_list_volume(self):
        group_id = '1234'
        grp = cs.groups.get(group_id, list_volume=True)
        cs.assert_called('GET', '/groups/%s?list_volume=True' % group_id)
        self._assert_request_id(grp)

    def test_create_group_from_src_snap(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.14'))
        grp = cs.groups.create_from_src('5678', None, name='group')
        expected = {
            'create-from-src': {
                'description': None,
                'name': 'group',
                'group_snapshot_id': '5678'
            }
        }
        cs.assert_called('POST', '/groups/action',
                         body=expected)
        self._assert_request_id(grp)

    def test_create_group_from_src_group_(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.14'))
        grp = cs.groups.create_from_src(None, '5678', name='group')
        expected = {
            'create-from-src': {
                'description': None,
                'name': 'group',
                'source_group_id': '5678'
            }
        }
        cs.assert_called('POST', '/groups/action',
                         body=expected)
        self._assert_request_id(grp)

    def test_enable_replication_group(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.38'))
        expected = {'enable_replication': {}}
        g0 = cs.groups.list()[0]
        grp = g0.enable_replication()
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.enable_replication('1234')
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.enable_replication(g0)
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)

    def test_disable_replication_group(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.38'))
        expected = {'disable_replication': {}}
        g0 = cs.groups.list()[0]
        grp = g0.disable_replication()
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.disable_replication('1234')
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.disable_replication(g0)
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)

    def test_failover_replication_group(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.38'))
        expected = {'failover_replication':
                    {'allow_attached_volume': False,
                     'secondary_backend_id': None}}
        g0 = cs.groups.list()[0]
        grp = g0.failover_replication()
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.failover_replication('1234')
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.failover_replication(g0)
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)

    def test_list_replication_targets(self):
        cs = fakes.FakeClient(api_versions.APIVersion('3.38'))
        expected = {'list_replication_targets': {}}
        g0 = cs.groups.list()[0]
        grp = g0.list_replication_targets()
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.list_replication_targets('1234')
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)
        grp = cs.groups.list_replication_targets(g0)
        self._assert_request_id(grp)
        cs.assert_called('POST', '/groups/1234/action', body=expected)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import api_versions
from cinderclient.tests.unit import utils
from cinderclient.tests.unit.v3 import fakes

defaults = fakes.FakeClient(api_versions.APIVersion('3.62'))


class VolumeTypeDefaultTest(utils.TestCase):

    def test_set(self):
        defaults.default_types.create('4c298f16-e339-4c80-b934-6cbfcb7525a0',
                                      '629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
        defaults.assert_called(
            'PUT', 'v3/default-types/629632e7-99d2-4c40-9ae3-106fa3b1c9b7',
            body={'default_type':
                  {'volume_type': '4c298f16-e339-4c80-b934-6cbfcb7525a0'}}
        )

    def test_get(self):
        defaults.default_types.list('629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
        defaults.assert_called(
            'GET', 'v3/default-types/629632e7-99d2-4c40-9ae3-106fa3b1c9b7')

    def test_get_all(self):
        defaults.default_types.list()
        defaults.assert_called(
            'GET', 'v3/default-types')

    def test_unset(self):
        defaults.default_types.delete('629632e7-99d2-4c40-9ae3-106fa3b1c9b7')
        defaults.assert_called(
            'DELETE', 'v3/default-types/629632e7-99d2-4c40-9ae3-106fa3b1c9b7')



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import base


class QuotaClassSet(base.Resource):

    @property
    def id(self):
        """Needed by base.Resource to self-refresh and be indexed."""
        return self.class_name

    def update(self, *args, **kwargs):
        return self.manager.update(self.class_name, *args, **kwargs)


class QuotaClassSetManager(base.Manager):
    resource_class = QuotaClassSet

    def get(self, class_name):
        return self._get("/os-quota-class-sets/%s" % (class_name),
                         "quota_class_set")

    def update(self, class_name, **updates):
        quota_class_set = {}

        for update in updates:
            quota_class_set[update] = updates[update]

        result = self._update('/os-quota-class-sets/%s' % (class_name),
                              {'quota_class_set': quota_class_set})
        return self.resource_class(self,
                                   result['quota_class_set'], loaded=True,
                                   resp=result.request_ids)



# Copyright (C) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""group snapshot interface (v3)."""


from cinderclient import api_versions
from cinderclient.apiclient import base as common_base
from cinderclient import base
from cinderclient import utils


class GroupSnapshot(base.Resource):
    """A group snapshot is a snapshot of a group."""
    def __repr__(self):
        return "<group_snapshot: %s>" % self.id

    def delete(self):
        """Delete this group snapshot."""
        return self.manager.delete(self)

    def update(self, **kwargs):
        """Update the name or description for this group snapshot."""
        return self.manager.update(self, **kwargs)

    def reset_state(self, state):
        """Reset the group snapshot's state with specified one."""
        return self.manager.reset_state(self, state)


class GroupSnapshotManager(base.ManagerWithFind):
    """Manage :class:`GroupSnapshot` resources."""
    resource_class = GroupSnapshot

    @api_versions.wraps('3.14')
    def create(self, group_id, name=None, description=None,
               user_id=None,
               project_id=None):
        """Creates a group snapshot.

        :param group_id: Name or uuid of a group
        :param name: Name of the group snapshot
        :param description: Description of the group snapshot
        :param user_id: User id derived from context
        :param project_id: Project id derived from context
        :rtype: :class:`GroupSnapshot`
       """

        body = {
            'group_snapshot': {
                'group_id': group_id,
                'name': name,
                'description': description,
            }
        }

        return self._create('/group_snapshots', body, 'group_snapshot')

    @api_versions.wraps('3.14')
    def get(self, group_snapshot_id):
        """Get a group snapshot.

        :param group_snapshot_id: The ID of the group snapshot to get.
        :rtype: :class:`GroupSnapshot`
        """
        return self._get("/group_snapshots/%s" % group_snapshot_id,
                         "group_snapshot")

    @api_versions.wraps('3.19')
    def reset_state(self, group_snapshot, state):
        """Update the provided group snapshot with the provided state.

        :param group_snapshot: The :class:`GroupSnapshot` to set the state.
        :param state: The state of the group snapshot to be set.
        """
        body = {'status': state} if state else {}
        return self._action('reset_status', group_snapshot, body)

    @api_versions.wraps('3.14')
    def list(self, detailed=True, search_opts=None):
        """Lists all group snapshots.

        :param detailed: list detailed info or not
        :param search_opts: search options
        :rtype: list of :class:`GroupSnapshot`
        """
        query_string = utils.build_query_param(search_opts, sort=True)

        detail = ""
        if detailed:
            detail = "/detail"

        return self._list("/group_snapshots%s%s" % (detail, query_string),
                          "group_snapshots")

    @api_versions.wraps('3.14')
    def delete(self, group_snapshot):
        """Delete a group_snapshot.

        :param group_snapshot: The :class:`GroupSnapshot` to delete.
        """
        return self._delete("/group_snapshots/%s" % base.getid(group_snapshot))

    @api_versions.wraps('3.14')
    def update(self, group_snapshot, **kwargs):
        """Update the name or description for a group_snapshot.

        :param group_snapshot: The :class:`GroupSnapshot` to update.
        """
        if not kwargs:
            return

        body = {"group_snapshot": kwargs}

        return self._update("/group_snapshots/%s" % base.getid(group_snapshot),
                            body)

    def _action(self, action, group_snapshot, info=None, **kwargs):
        """Perform a group_snapshot action."""
        body = {action: info}
        self.run_hooks('modify_body_for_action', body, **kwargs)
        url = '/group_snapshots/%s/action' % base.getid(group_snapshot)
        resp, body = self.api.client.post(url, body=body)
        return common_base.TupleWithMeta((resp, body), resp)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import io
import sys
from unittest import mock

import ddt

from cinderclient import api_versions
from cinderclient.apiclient import base as common_base
from cinderclient import base
from cinderclient import exceptions
from cinderclient import shell_utils
from cinderclient.tests.unit import utils as test_utils
from cinderclient import utils

REQUEST_ID = 'req-test-request-id'
UUID = '8e8ec658-c7b0-4243-bdf8-6f7f2952c0d0'


class FakeResource(object):
    NAME_ATTR = 'name'

    def __init__(self, _id, properties, **kwargs):
        self.id = _id
        try:
            self.name = properties['name']
        except KeyError:
            pass

    def append_request_ids(self, resp):
        pass


class FakeManager(base.ManagerWithFind):

    resource_class = FakeResource

    resources = [
        FakeResource('1234', {'name': 'entity_one'}),
        FakeResource(UUID, {'name': 'entity_two'}),
        FakeResource('5678', {'name': '9876'})
    ]

    def get(self, resource_id, **kwargs):
        for resource in self.resources:
            if resource.id == str(resource_id):
                return resource
        raise exceptions.NotFound(resource_id)

    def list(self, search_opts, **kwargs):
        return common_base.ListWithMeta(self.resources, REQUEST_ID)


class FakeManagerWithApi(base.Manager):

    @api_versions.wraps('3.1')
    def return_api_version(self):
        return '3.1'

    @api_versions.wraps('3.2')
    def return_api_version(self):  # noqa: F811
        return '3.2'


class FakeDisplayResource(object):
    NAME_ATTR = 'display_name'

    def __init__(self, _id, properties):
        self.id = _id
        try:
            self.display_name = properties['display_name']
        except KeyError:
            pass

    def append_request_ids(self, resp):
        pass


class FakeDisplayManager(FakeManager):

    resource_class = FakeDisplayResource

    resources = [
        FakeDisplayResource('4242', {'display_name': 'entity_three'}),
    ]


class FindResourceTestCase(test_utils.TestCase):

    def setUp(self):
        super(FindResourceTestCase, self).setUp()
        self.manager = FakeManager(None)

    def test_find_none(self):
        self.manager.find = mock.Mock(side_effect=self.manager.find)
        self.assertRaises(exceptions.CommandError,
                          utils.find_resource,
                          self.manager,
                          'asdf')
        self.assertEqual(2, self.manager.find.call_count)

    def test_find_by_integer_id(self):
        output = utils.find_resource(self.manager, 1234)
        self.assertEqual(self.manager.get('1234'), output)

    def test_find_by_str_id(self):
        output = utils.find_resource(self.manager, '1234')
        self.assertEqual(self.manager.get('1234'), output)

    def test_find_by_uuid(self):
        output = utils.find_resource(self.manager, UUID)
        self.assertEqual(self.manager.get(UUID), output)

    def test_find_by_str_name(self):
        output = utils.find_resource(self.manager, 'entity_one')
        self.assertEqual(self.manager.get('1234'), output)

    def test_find_by_str_displayname(self):
        display_manager = FakeDisplayManager(None)
        output = utils.find_resource(display_manager, 'entity_three')
        self.assertEqual(display_manager.get('4242'), output)

    def test_find_by_group_id(self):
        output = utils.find_resource(self.manager, 1234, is_group=True,
                                     list_volume=True)
        self.assertEqual(self.manager.get('1234', list_volume=True), output)

    def test_find_by_group_name(self):
        display_manager = FakeDisplayManager(None)
        output = utils.find_resource(display_manager, 'entity_three',
                                     is_group=True, list_volume=True)
        self.assertEqual(display_manager.get('4242', list_volume=True),
                         output)


class CaptureStdout(object):
    """Context manager for capturing stdout from statements in its block."""
    def __enter__(self):
        self.real_stdout = sys.stdout
        self.stringio = io.StringIO()
        sys.stdout = self.stringio
        return self

    def __exit__(self, *args):
        sys.stdout = self.real_stdout
        self.stringio.seek(0)
        self.read = self.stringio.read


@ddt.ddt
class BuildQueryParamTestCase(test_utils.TestCase):

    def test_build_param_without_sort_switch(self):
        dict_param = {
            'key1': 'val1',
            'key2': 'val2',
            'key3': 'val3',
        }
        result = utils.build_query_param(dict_param, True)

        self.assertIn('key1=val1', result)
        self.assertIn('key2=val2', result)
        self.assertIn('key3=val3', result)

    def test_build_param_with_sort_switch(self):
        dict_param = {
            'key1': 'val1',
            'key2': 'val2',
            'key3': 'val3',
        }
        result = utils.build_query_param(dict_param, True)

        expected = "?key1=val1&key2=val2&key3=val3"
        self.assertEqual(expected, result)

    @ddt.data({},
              None,
              {'key1': 'val1', 'key2': None, 'key3': False, 'key4': ''})
    def test_build_param_with_nones(self, dict_param):
        result = utils.build_query_param(dict_param)

        expected = ("key1=val1", "key3=False") if dict_param else ()
        for exp in expected:
            self.assertIn(exp, result)
        if not expected:
            self.assertEqual("", result)


@ddt.ddt
class ExtractFilterTestCase(test_utils.TestCase):

    @ddt.data({'content': ['key1=value1'],
               'expected': {'key1': 'value1'}},
              {'content': ['key1={key2:value2}'],
               'expected': {'key1': {'key2': 'value2'}}},
              {'content': ['key1=value1', 'key2={key22:value22}'],
               'expected': {'key1': 'value1', 'key2': {'key22': 'value22'}}})
    @ddt.unpack
    def test_extract_filters(self, content, expected):
        result = shell_utils.extract_filters(content)
        self.assertEqual(expected, result)


class PrintListTestCase(test_utils.TestCase):

    def test_print_list_with_list(self):
        Row = collections.namedtuple('Row', ['a', 'b'])
        to_print = [Row(a=3, b=4), Row(a=1, b=2)]
        with CaptureStdout() as cso:
            shell_utils.print_list(to_print, ['a', 'b'])
        # Output should be sorted by the first key (a)
        self.assertEqual("""\
+---+---+
| a | b |
+---+---+
| 1 | 2 |
| 3 | 4 |
+---+---+
""", cso.read())

    def test_print_list_with_None_data(self):
        Row = collections.namedtuple('Row', ['a', 'b'])
        to_print = [Row(a=3, b=None), Row(a=1, b=2)]
        with CaptureStdout() as cso:
            shell_utils.print_list(to_print, ['a', 'b'])
        # Output should be sorted by the first key (a)
        self.assertEqual("""\
+---+---+
| a | b |
+---+---+
| 1 | 2 |
| 3 | - |
+---+---+
""", cso.read())

    def test_print_list_with_list_sortby(self):
        Row = collections.namedtuple('Row', ['a', 'b'])
        to_print = [Row(a=4, b=3), Row(a=2, b=1)]
        with CaptureStdout() as cso:
            shell_utils.print_list(to_print, ['a', 'b'], sortby_index=1)
        # Output should be sorted by the second key (b)
        self.assertEqual("""\
+---+---+
| a | b |
+---+---+
| 2 | 1 |
| 4 | 3 |
+---+---+
""", cso.read())

    def test_print_list_with_list_no_sort(self):
        Row = collections.namedtuple('Row', ['a', 'b'])
        to_print = [Row(a=3, b=4), Row(a=1, b=2)]
        with CaptureStdout() as cso:
            shell_utils.print_list(to_print, ['a', 'b'], sortby_index=None)
        # Output should be in the order given
        self.assertEqual("""\
+---+---+
| a | b |
+---+---+
| 3 | 4 |
| 1 | 2 |
+---+---+
""", cso.read())

    def test_print_list_with_generator(self):
        Row = collections.namedtuple('Row', ['a', 'b'])

        def gen_rows():
            for row in [Row(a=1, b=2), Row(a=3, b=4)]:
                yield row
        with CaptureStdout() as cso:
            shell_utils.print_list(gen_rows(), ['a', 'b'])
        self.assertEqual("""\
+---+---+
| a | b |
+---+---+
| 1 | 2 |
| 3 | 4 |
+---+---+
""", cso.read())

    def test_print_list_with_return(self):
        Row = collections.namedtuple('Row', ['a', 'b'])
        to_print = [Row(a=3, b='a\r'), Row(a=1, b='c\rd')]
        with CaptureStdout() as cso:
            shell_utils.print_list(to_print, ['a', 'b'])
        # Output should be sorted by the first key (a)
        self.assertEqual("""\
+---+-----+
| a | b   |
+---+-----+
| 1 | c d |
| 3 | a   |
+---+-----+
""", cso.read())


class PrintDictTestCase(test_utils.TestCase):

    def test__pretty_format_dict(self):
        content = {'key1': 'value1', 'key2': 'value2'}
        expected = "key1 : value1\nkey2 : value2"
        result = shell_utils._pretty_format_dict(content)
        self.assertEqual(expected, result)

    def test_print_dict_with_return(self):
        d = {'a': 'A', 'b': 'B', 'c': 'C', 'd': 'test\rcarriage\n\rreturn'}
        with CaptureStdout() as cso:
            shell_utils.print_dict(d)
        self.assertEqual("""\
+----------+---------------+
| Property | Value         |
+----------+---------------+
| a        | A             |
| b        | B             |
| c        | C             |
| d        | test carriage |
|          |  return       |
+----------+---------------+
""", cso.read())

    def test_print_dict_with_dict_inside(self):
        content = {'a': 'A', 'b': 'B', 'f_key':
                   {'key1': 'value1', 'key2': 'value2'}}
        with CaptureStdout() as cso:
            shell_utils.print_dict(content, formatters='f_key')
        self.assertEqual("""\
+----------+---------------+
| Property | Value         |
+----------+---------------+
| a        | A             |
| b        | B             |
| f_key    | key1 : value1 |
|          | key2 : value2 |
+----------+---------------+
""", cso.read())
